from enum import Enum


class Status(str, Enum):
    NOT_ASSIGNED = 0
    READY_TO_DEPLOY = 1
    ACTIVE_ON_WORKFLOW = 2
    INACTIVE_ON_WORKFLOW = 3 # inactive customer
    EXIT_WITH_CONVERSION = 4
    SUSPENDED = 5



from unittest.mock import patch
from typing import Callable

from versionhq.agent.model import Agent, LLM
from versionhq.task.model import Task, ResponseField, TaskOutput, TaskExecutionType
from versionhq.tool.model import Tool, ToolSet
from versionhq.tool.decorator import tool


# def test_async_execute_task():
#     task = Task(description="Return string: 'test'", type=TaskExecutionType.ASYNC)

#     with patch.object(Agent, "execute_task", return_value="test") as execute:
#         res = task.execute()
#         assert res.raw == "test"
#         execute.assert_called_once_with(task=task, context=None, task_tools=list())


# def test_sync_execute_with_task_context():
#     """
#     Use case = One agent handling multiple tasks sequentially using context set in the main task.
#     """
#     sub_task = Task(
#         description="return the output following the given prompt.",
#         response_fields=[
#             ResponseField(title="subtask_result", data_type=str, required=True),
#         ]
#     )
#     main_task = Task(
#         description="return the output following the given prompt.",
#         response_fields=[
#             ResponseField(title="test1", data_type=int, required=True),
#             ResponseField(title="test2", data_type=str, required=True),
#         ],
#     )
#     res = main_task.execute(context=[sub_task])

#     assert isinstance(res, TaskOutput)
#     assert res.task_id is main_task.id
#     assert res.raw is not None
#     assert isinstance(res.raw, str)
#     assert res.json_dict is not None
#     assert isinstance(res.json_dict, dict)
#     assert res.pydantic is None
#     assert sub_task.output is not None
#     assert sub_task.output.json_dict is not None
#     assert "subtask_result" in main_task._prompt()



# def test_callback():
#     """
#     See if the callback function is executed well with kwargs.
#     """

#     def callback_func(condition: str, test1: str):
#         # task_id = str(id) if id else None
#         return f"Result: {test1}, condition added: {condition}"

#     task = Task(
#         description="return the output following the given prompt.",
#         response_fields=[
#             ResponseField(title="test1", data_type=str, required=True),
#         ],
#         callback=callback_func,
#         callback_kwargs=dict(condition="demo for pytest")
#     )
#     res = task.execute()

#     assert res and isinstance(res, TaskOutput)
#     assert res.task_id is task.id
#     assert "demo for pytest" in res.callback_output


# def test_delegate():
#     # agent = Agent(role="demo agent 6", goal="My amazing goals", maxit=1, max_tokens=3000)
#     task = Task(
#         description="return the output following the given prompt.",
#         response_fields=[
#             ResponseField(title="test1", data_type=str, required=True),
#         ],
#         allow_delegation=True
#     )
#     task.execute()

#     assert task.output is not None
#     assert "vhq-Delegated-Agent" in task.processed_agents
#     assert task.delegations != 0


# def test_conditional_task():
#     task = Task(
#         description="erturn the output following the given prompt.",
#         response_fields=[ResponseField(title="test1", data_type=str, required=True),],
#     )
#     res = task.execute_sync(agent=base_agent)

#     conditional_task = ConditionalTask(
#         description="return the output following the given prompt.",
#         response_fields=[ResponseField(title="test1", data_type=str, required=True),],
#         condition=lambda x: bool("zzz" in task.output.raw)
#     )
#     should_execute = conditional_task.should_execute(context=res)

#     assert res.raw is not None
#     assert should_execute is False

#     conditional_res = conditional_task._handle_conditional_task(task_outputs=[res,], task_index=1, was_replayed=False)
#     if not should_execute:
#         assert conditional_res is None
#     else:
#         assert conditional_res.task_id is conditional_task.id


def test_store_task_log():
    task = Task(
        description="return the output following the given prompt.",
        response_fields=[ResponseField(title="task_1", data_type=str, required=True),],
    )
    assert task._task_output_handler.load() is not None


def test_task_with_agent_tools():
    simple_tool = Tool(name="simple tool", func=lambda x: "simple func")
    agent = Agent(role="demo", goal="execute tools", tools=[simple_tool,], maxit=1, max_tokens=3000)
    task = Task(description="execute tool", can_use_agent_tools=True, tool_res_as_final=True)
    res = task.execute(agent=agent)
    assert res.tool_output == "simple func"

    def empty_func():
        return "empty func"

    func_tool = Tool(name="func tool", func=empty_func)
    agent.tools = [func_tool]
    res = task.execute(agent=agent)
    assert res.tool_output == "empty func"


    def demo_func(message: str) -> str:
        return message + "_demo"

    class CustomTool(Tool):
        name: str = "custom tool"

    custom_tool = CustomTool(func=demo_func)
    agent.tools = [custom_tool]
    res = task.execute(agent=agent)
    assert "_demo" in res.tool_output


def test_task_with_tools():

    def random_func(message: str) -> str:
        return message + "_demo"

    tool = Tool(name="tool", func=random_func)
    tool_set = ToolSet(tool=tool, kwargs=dict(message="empty func"))
    task = Task(description="execute the given tools", tools=[tool_set,], tool_res_as_final=True)
    # res = task.execute()
    # assert res.tool_output == "empty func_demo"

    class CustomTool(Tool):
        name: str = "custom tool"
        func: Callable = None

    custom_tool = CustomTool(func=random_func)
    task.tools = [custom_tool]
    res = task.execute()
    assert "_demo" in res.tool_output

    task.tools = [custom_tool]
    res = task.execute()
    assert res.tool_output is not None



def test_task_without_response_format():
    task = Task(description="return a simple output with any random values.")
    res = task.execute()

    assert res and isinstance(res, TaskOutput)
    assert res.json_dict and isinstance(res.json_dict, dict)
    assert res.pydantic is None


def test_build_agent_without_developer_prompt():
    agent = Agent(
        role="analyst",
        goal="analyze the company's website and retrieve the product overview",
        backstory="You are competitive analysts who have abundand knowledge in marketing, product management.",
        use_developer_prompt=False
    )
    task = Task(description="return a simple output with any random values.")
    res = task.execute(agent=agent)

    assert res and isinstance(res, TaskOutput)
    assert res.json_dict and isinstance(res.json_dict, dict)
    assert res.pydantic is None


# def test_callback_with_custom_output():
#     class CustomOutput(BaseModel):
#         test1: str
#         test2: list[str]

#     def dummy_func(message: str, test1: str, test2: list[str]) -> str:
#         return f"""{message}: {test1}, {", ".join(test2)}"""

#     task = Task(
#         description="Amazing task",
#         pydantic_output=CustomOutput,
#         callback=dummy_func,
#         callback_kwargs=dict(message="Hi! Here is the result: ")
#     )
#     res = task.execute(context="amazing context to consider.")

#     assert res.task_id == task.id
#     assert res.pydantic.test1 and res.pydantic.test2
#     assert "Hi! Here is the result: " in res.callback_output and res.pydantic.test1 in res.callback_output and ", ".join(res.pydantic.test2) in res.callback_output


def test_task_with_agent_callback():
    import litellm

    def dummy_func(*args, **kwargs) -> str:
        return "Demo func"

    agent = Agent(role="demo", goal="amazing project goal", maxit=1, max_tokens=3000, callbacks=[dummy_func,])
    task = Task(description="Amazing task")
    res = task.execute(agent=agent)

    assert res.raw and res.task_id == task.id
    assert litellm.callbacks == [dummy_func]


def test_rpm():
    agent = Agent(role="demo", goal="use the given tools", max_tokens=3000, max_rpm=3)
    assert agent._rpm_controller and agent._rpm_controller.max_rpm == 3

    a = "hello"
    tool = Tool(func=lambda x: a + x)
    tool_set = ToolSet(tool=tool, kwargs={ "x": "_demo" })
    task = Task(description="Summarize overview of the given tool in sentence, then execute the tool.", tools=[tool_set,])
    res = task.execute(agent=agent)

    assert "hello_demo" in res.raw if res.raw else res
    assert agent._rpm_controller._current_rpm < agent._rpm_controller.max_rpm


def test_maxit():
    @tool
    def demo() -> str:
        """Get the final answer but don't give it yet, just re-use this
        tool non-stop."""
        return "demo"

    agent = Agent(role="demo", goal="amazing demo", maxit=2)
    task = Task(description="Summarize overview of the given tool in sentences.", tools=[demo,])

    with patch.object(LLM, "call", wraps=agent.llm.call) as mock:
        task.execute(agent=agent)
        assert mock.call_count <= 2


# def test_evaluation():
#     """
#     See if the output will be evaluated accurately - when the task was given eval criteria
#     """
#     from versionhq.task.evaluate import Evaluation, EvaluationItem
#     from versionhq.agent.inhouse_agents import vhq_task_evaluator

#     task = Task(
#         description="Research a topic to teach a kid aged 6 about math.",
#         should_evaluate=True,
#         eval_criteria=["Uniquness of the topic researched", "Fit to the target audience",],
#     )
#     res = task.execute()

#     assert res.evaluation and isinstance(res.evaluation, Evaluation)
#     assert [isinstance(item, EvaluationItem) and item.criteria in task.eval_criteria for item in res.evaluation.items]
#     assert res.evaluation.latency and res.evaluation.tokens and res.evaluation.eval_by == vhq_task_evaluator
#     assert res.evaluation.aggregate_score is not None and res.evaluation.suggestion_summary



import matplotlib
matplotlib.use('agg')

import enum
import uuid
import networkx as nx
import matplotlib.pyplot as plt
from abc import ABC
from typing import List, Any, Optional, Callable, Dict, Type, Tuple

from pydantic import BaseModel, InstanceOf, Field, UUID4, field_validator
from pydantic_core import PydanticCustomError

from versionhq.task.model import Task, TaskOutput
from versionhq.agent.model import Agent
from versionhq._utils.logger import Logger


class TaskStatus(enum.Enum):
    """
    Enum to track the task execution status
    """
    NOT_STARTED = 1
    IN_PROGRESS = 2
    WAITING = 3     # waiting for its dependant tasks to complete. resumption set as AUTO.
    COMPLETED = 4
    DELAYED = 5     # task in progress - but taking longer than expected duration
    ON_HOLD = 6     # intentionally paused due to external factors and decisions. resumption set as DECISION.
    ERROR = 7       # tried task execute but returned error. resupmtion follows edge weights and agent settings


class DependencyType(enum.Enum):
    """
    Concise enumeration of the edge type.
    """
    FINISH_TO_START = "FS"  # Task B starts after Task A finishes
    START_TO_START = "SS"  # Task B starts when Task A starts
    FINISH_TO_FINISH = "FF"  # Task B finishes when Task A finishes
    START_TO_FINISH = "SF"  # Task B finishes when Task A starts


class Node(BaseModel):
    """
    A class to store a node object.
    """

    id: UUID4 = Field(default_factory=uuid.uuid4, frozen=True)
    task: InstanceOf[Task] = Field(default=None)
    in_degree_nodes: List[Any] = Field(default_factory=list, description="list of Node objects")
    out_degree_nodes: List[Any] = Field(default_factory=list, description="list of Node objects")
    assigned_to: InstanceOf[Agent] = Field(default=None)
    status: TaskStatus = Field(default=TaskStatus.NOT_STARTED)


    @field_validator("id", mode="before")
    @classmethod
    def _deny_user_set_id(cls, v: Optional[UUID4]) -> None:
        if v:
            raise PydanticCustomError("may_not_set_field", "This field is not to be set by client.", {})


    def is_independent(self) -> bool:
        return not self.in_degree_nodes and not self.out_degree_nodes

    def handle_task_execution(self, agent: Agent = None, context: str = None, format: Type[BaseModel] = None) -> TaskOutput | None:
        """
        Start task execution and update status accordingly.
        """

        self.status = TaskStatus.IN_PROGRESS

        if not self.task:
            Logger(verbose=True).log(level="error", message="Missing a task to execute. We'll return None.", color="red")
            self.status = TaskStatus.ERROR
            return None

        self.task.pydantic_output = self.task.pydantic_output if self.task.pydantic_output else format if type(format) == BaseModel else None
        res = self.task.execute(agent=agent, context=context)
        self.status = TaskStatus.COMPLETED if res else TaskStatus.ERROR
        return res


    @property
    def in_degrees(self) -> int:
        return len(self.in_degree_nodes) if self.in_degree_nodes else 0

    @property
    def out_degrees(self) -> int:
        return len(self.out_degree_nodes) if self.out_degree_nodes else 0

    @property
    def degrees(self) -> int:
        return self.in_degrees + self.out_degrees

    @property
    def identifier(self) -> str:
        """Unique identifier of the node"""
        return f"{str(self.id)}"

    def __str__(self):
        if self.task:
            return f"{self.identifier}: {self.task.name if self.task.name else self.task.description[0: 12]}"
        else:
            return self.identifier


class Edge(BaseModel):
    """
    A class to store an edge object that connects source and target nodes.
    """

    source: Node = Field(default=None)
    target: Node = Field(default=None)

    description: Optional[str] = Field(default=None)
    weight: Optional[float | int] = Field(default=1, description="est. duration for the task execution or respective weight of the target node (1 low - 10 high priority)")

    dependency_type: DependencyType = Field(default=DependencyType.FINISH_TO_START)
    required: bool = Field(default=True, description="whether to consider the source's status")
    condition: Optional[Callable] = Field(default=None, description="conditional function to start executing the dependency")
    condition_kwargs: Optional[Dict[str, Any]] = Field(default_factory=dict)

    lag: Optional[float | int] = Field(default=None, description="lag time (sec) from the dependency met to the task execution")
    data_transfer: bool = Field(default=True, description="whether the data transfer is required. by default transfer plane text output from in-degree nodes as context")


    def _response_schema(self) -> Type[BaseModel]:
        class EdgeResponseSchema(BaseModel):
            wight: float
            dependecy_type: str
            required: bool
            need_condition: bool
            lag_in_sec: float

        return EdgeResponseSchema


    def dependency_met(self) -> bool:
        """
        Defines if the dependency is ready to execute:

        required - condition - Dependency Met? - Dependent Task Can Start?
        True	Not Given	Predecessor task finished	Yes
        True	Given	Predecessor task finished and condition True	Yes
        False	Not Given	Always (regardless of predecessor status)	Yes
        False	Given	Condition True (predecessor status irrelevant)	Yes
        """

        if not self.required:
            return self.condition(**self.condition_kwargs) if self.condition else True

        match self.dependency_type:
            case DependencyType.FINISH_TO_START:
                """target starts after source finishes"""
                if not self.source or self.source.status == TaskStatus.COMPLETED:
                    return self.condition(**self.conditon_kwargs) if self.condition else True
                else:
                    return False

            case DependencyType.START_TO_START:
                """target starts when source starts"""
                if not self.source or self.source.status != TaskStatus.NOT_STARTED:
                    return self.condition(**self.conditon_kwargs) if self.condition else True
                else:
                    return False

            case DependencyType.FINISH_TO_FINISH:
                """target finish when source start"""
                if not self.source or self.source.status != TaskStatus.COMPLETED:
                    return self.condition(**self.conditon_kwargs) if self.condition else True
                else:
                    return False

            case DependencyType.START_TO_FINISH:
                """target finishes when source start"""
                if not self.source or self.source.status == TaskStatus.IN_PROGRESS:
                    return self.condition(**self.conditon_kwargs) if self.condition else True
                else:
                    return False


    def activate(self, format: Type[BaseModel] = None) -> TaskOutput | None:
        """
        Activates the edge to initiate task execution of the target node.
        """

        if not self.source or not self.target:
            Logger(verbose=True).log(level="warning", message="Cannot find source or target nodes. We'll return None.", color="yellow")
            return None

        if not self.dependency_met():
            Logger(verbose=True).log(level="warning", message="Dependencies not met. We'll return None.", color="yellow")
            return None

        if self.lag:
            import time
            time.sleep(self.lag)

        context = self.source.task.output.raw if self.data_transfer else None
        res = self.target.handle_task_execution(context=context, format=format)
        return res


class Graph(ABC, BaseModel):
    """
    An abstract class to store G using NetworkX library.
    """
    directed: bool = Field(default=False, description="Whether the graph is directed")
    graph: Type[nx.Graph] = Field(default=None)
    nodes: Dict[str, Node] = Field(default_factory=dict, description="identifier: Node - for the sake of ")
    edges: Dict[str, Edge] = Field(default_factory=dict)

    def __init__(self, directed: bool = False, **kwargs):
        super().__init__(directed=directed, **kwargs)
        self.graph = nx.DiGraph(directed=True) if self.directed else nx.Graph()


    def _return_node_object(self, node_identifier) -> Node | None:
        match = [v for k, v in self.nodes.items() if k == node_identifier]

        if match:
            node = match[0] if isinstance(match[0], Node) else match[0]["node"] if "node" in match[0] else None
            return node
        else:
            return None

    def add_node(self, node: Node) -> None:
        if node.identifier in self.nodes.keys():
            return
        self.graph.add_node(node.identifier, node=node)
        self.nodes[node.identifier] = node

    def add_edge(self, source: str, target: str, edge: Edge) -> None:
        self.graph.add_edge(source, target, edge=edge)

        source_node, target_node = self._return_node_object(source), self._return_node_object(target)

        edge.source = source_node
        source_node.out_degree_nodes.append(target_node)

        edge.target = target_node
        target_node.in_degree_nodes.append(source_node)

        self.edges[(source, target)] = edge

    def add_weighted_edges_from(self, edges):
        self.graph.add_weighted_edges_from(edges)

    def get_neighbors(self, node: Node) -> List[Node]:
        return list(self.graph.neighbors(node))

    def get_in_degree(self, node: Node) -> int:
        return self.graph.in_degree(node)

    def get_out_degree(self, node: Node) -> int:
        return self.graph.out_degree(node)

    def find_start_nodes(self) -> List[Node]:
        return [v for v in self.nodes.values() if v.in_degrees == 0 and v.out_degrees > 0]

    def find_end_nodes(self) ->  List[Node]:
        return [v for v in self.nodes.values() if v.out_degrees == 0 and v.in_degrees > 0]

    def find_critical_end_node(self) -> Node | None:
        """
        Find a critical end node from all the end nodes to lead a conclusion of the entire graph.
        """
        end_nodes = self.find_end_nodes()
        if not end_nodes:
            return None

        if len(end_nodes) == 1:
            return end_nodes[0]

        edges = [v for k, v in self.edges if isinstance(v, Edge) and v.source in end_nodes]
        critical_edge = max(edges, key=lambda item: item['weight']) if edges else None
        return critical_edge.target if critical_edge else None


    def find_path(self, source: Optional[str] | None, target: str, weight: Optional[Any] | None) -> Any:
        try:
            return nx.shortest_path(self.graph, source=source, target=target, weight=weight)
        except nx.NetworkXNoPath:
            return None

    def find_all_paths(self,  source: str, target: str) -> List[Any]:
        return list(nx.all_simple_paths(self.graph, source=source, target=target))

    def find_critical_path(self) -> tuple[List[Any], int, Dict[str, int]]:
        """
        Finds the critical path in the graph.
        Returns:
            A tuple containing:
                - The critical path (a list of edge identifiers).
                - The duration of the critical path.
                - A dictionary of all paths and their durations.
        """

        all_paths = {}
        for start_node in self.find_start_nodes():
            for end_node in self.find_end_nodes(): # End at nodes with 0 out-degree
                for edge in nx.all_simple_paths(self.graph, source=start_node.identifier, target=end_node.identifier):
                    edge_weight = sum(self.edges.get(item).weight if self.edges.get(item) else 0 for item in edge)
                    all_paths[tuple(edge)] = edge_weight

        if not all_paths:
            return [], 0, all_paths

        critical_path = max(all_paths, key=all_paths.get)
        critical_duration = all_paths[critical_path]

        return list(critical_path), critical_duration, all_paths


    def is_circled(self, node: Node) -> bool:
        """Check if there's a path from the node to itself and return bool."""
        try:
            path = nx.shortest_path(self.graph, source=node, target=node)
            return True if path else False
        except nx.NetworkXNoPath:
            return False


class TaskGraph(Graph):
    id: UUID4 = Field(default_factory=uuid.uuid4, frozen=True)
    should_reform: bool = Field(default=False)
    outputs: Dict[str, TaskOutput] = Field(default_factory=dict, description="stores node identifier and TaskOutput")
    concl_template: Optional[Dict[str, Any] | Type[BaseModel]] = Field(default=None, description="stores a format of `concl` either in Pydantic model or JSON dict")
    concl: Any = Field(default=None, description="store the final result of the entire task graph")


    def _save(self, title: str, abs_file_path: str = None) -> None:
        """
        Save the graph image in the local directory.
        """
        try:
            import os
            project_root = os.path.abspath(os.getcwd())
            abs_file_path = abs_file_path if abs_file_path else f"{project_root}/.diagrams"
            title = title if title else f"vhq-Diagram-{str(self.id)}"

            os.makedirs(abs_file_path, exist_ok=True)
            plt.savefig(f"{abs_file_path}/{title}.png")

        except Exception as e:
            Logger().log(level="error", message=f"Failed to save the graph {str(self.id)}: {str(e)}", color="red")


    def add_task(self, task: Node | Task) -> Node:
        """Convert `task` to a Node object and add it to G"""

        if isinstance(task, Node) and task.identifier in self.nodes.keys():
            return task

        elif isinstance(task, Task):
            match = []
            for v in self.nodes.values():
                if type(v) == dict and v["node"] and v["node"].task == task:
                    match.append(v["node"])
                elif v.task == task:
                    match.append(v)

            if match:
                return match[0]
            else:
                node = Node(task=task)
                self.add_node(node)
                return node

        else:
            task_node = task if isinstance(task, Node) else Node(task=task)
            self.add_node(task_node)
            return task_node


    def add_dependency(self, source_node_identifier: str, target_node_identifier: str, **edge_attributes) -> None:
        """
        Add an edge that connect task 1 (source) and task 2 (target) using task_node.name as an identifier
        """

        if not edge_attributes:
            Logger().log(level="error", message="Edge attributes are missing.", color="red")

        edge = Edge()
        for k in Edge.model_fields.keys():
            v = edge_attributes.get(k, None)
            if v:
                setattr(edge, k, v)
            else:
                pass

        self.add_edge(source_node_identifier, target_node_identifier, edge)


    def get_task_status(self, identifier: str) -> TaskStatus | None:
        """Retrieves the latest status of the given node"""
        if not identifier or identifier not in self.nodes.keys():
            Logger().log(level="error", message=f"Task node: {identifier} is not in the graph.", color="red")
            return None

        return self._return_node_object(identifier).status


    def visualize(self, layout: str = None, should_save: bool = False):
        from matplotlib.lines import Line2D
        from versionhq.task_graph.colors import white, black, darkgrey, grey, primary, orange, lightgreen, green, darkgreen, darkergreen

        try:
            pos = nx.drawing.nx_agraph.graphviz_layout(self.graph, prog='dot') # 'dot', 'neato', 'fdp', 'sfdp'
        except ImportError:
            pos = nx.spring_layout(self.graph, seed=42) # REFINEME - layout

        node_colors, legend_elements = list(), list()
        for k, v in self.nodes.items():
            status = self.get_task_status(identifier=k)
            node =  v if isinstance(v, Node) else v["node"] if hasattr(v, "node") else None
            if node:
                task_name = node.task.name if node.task and node.task.name else node.task.description if node.task else None
                legend_label = f"ID {str(node.identifier)[0: 6]}...: {task_name}" if task_name else f"ID {str(node.identifier)[0: 6]}..."
                match status:
                    case TaskStatus.NOT_STARTED:
                        node_colors.append(green)
                        legend_elements.append(Line2D([0], [0], marker='o', color='w', label=legend_label, markerfacecolor=green))

                    case TaskStatus.IN_PROGRESS:
                        node_colors.append(darkgreen)
                        legend_elements.append(Line2D([0], [0], marker='o', color='w', label=legend_label, markerfacecolor=darkgreen))

                    case TaskStatus.WAITING:
                        node_colors.append(lightgreen)
                        legend_elements.append(Line2D([0], [0], marker='o', color='w', label=legend_label, markerfacecolor=lightgreen))

                    case TaskStatus.COMPLETED:
                        node_colors.append(darkergreen)
                        legend_elements.append(Line2D([0], [0], marker='o', color='w', label=legend_label, markerfacecolor=darkergreen))

                    case TaskStatus.DELAYED:
                        node_colors.append(orange)
                        legend_elements.append(Line2D([0], [0], marker='o', color='w', label=legend_label, markerfacecolor=orange))

                    case TaskStatus.ON_HOLD:
                        node_colors.append(grey)
                        legend_elements.append(Line2D([0], [0], marker='o', color='w', label=legend_label, markerfacecolor=grey))

                    case TaskStatus.ERROR:
                        node_colors.append(primary)
                        legend_elements.append(Line2D([0], [0], marker='o', color='w', label=legend_label, markerfacecolor=primary))

                    case _:
                        node_colors.append(white)
                        legend_elements.append(Line2D([0], [0], marker='o', color='w', label=legend_label, markerfacecolor=white))

        critical_paths, duration, paths = self.find_critical_path()
        edge_colors = [black if (u, v) in zip(critical_paths, critical_paths[1:]) else darkgrey for u, v in self.graph.edges()]
        edge_widths = [v.weight * 0.5 for v in self.edges.values()]

        # for k, v in self.edges.items():
        #     # edge_weights = nx.get_edge_attributes(self.graph, 'weight')
        #     # edge_colors.append(plt.cm.viridis(v.weight / max(edge_weights.values())))
        #     edge_widths.append(v.weight * 0.5)

        nx.draw(
            self.graph, pos,
            with_labels=True,
            node_size=600,
            node_color=node_colors,
            font_size=8,
            font_color=darkgrey,
            edge_color=edge_colors,
            width=edge_widths,
            arrows=True,
            arrowsize=15,
            arrowstyle='-|>'
        )

        edge_labels = nx.get_edge_attributes(G=self.graph, name="edges")
        nx.draw_networkx_edge_labels(self.graph, pos, edge_labels=edge_labels)

        plt.legend(handles=legend_elements, loc='lower right')
        plt.title(f"vhq-Diagram {str(self.id)}")

        if should_save:
            self._save(title=f"vhq-Diagram {str(self.id)}")
        plt.show(block=False)


    def activate(self, target_node_identifier: Optional[str] = None) -> Tuple[TaskOutput | None, Dict[str, TaskOutput]]:
        """
        Starts to execute all nodes in the graph or a specific node if the target is given, following the given conditons of the edge obeject.
        Then returns tuple of the last task output and all task outputs (self.outputs)
        """

        Logger().log(color="blue", message=f"Start to activate the graph: {str(self.id)}", level="info")

        if target_node_identifier:
            if not [k for k in self.nodes.keys() if k == target_node_identifier]:
                Logger().log(level="error", message=f"The node {str(target_node_identifier)} is not in the graph.", color="red")
                return None

            # find a shortest path to each in-degree node of the node and see if dependency met.
            node = self._return_node_object(target_node_identifier)
            sources = node.in_degrees
            edge_status = []
            res = None

            for item in sources:
                edge = self.find_path(source=item, target=target_node_identifier)
                edge_status.append(dict(edge=edge if edge else None, dep_met=edge.dependency_met() if edge else False))

            if len([item for item in edge_status if item["dep_met"] == True]) == len(sources):
                res = node.handle_task_execution()
                self.outputs.update({ target_node_identifier: res })

            return res, self.outputs

        else:
            if not self.edges or not self.nodes:
                Logger().log(level="error", message="TaskGraph needs at least 2 nodes and 1 edge to activate. We'll return None.", color="red")
                return None

            start_nodes = self.find_start_nodes()
            end_nodes = self.find_end_nodes()
            critical_end_node = self.find_critical_end_node()
            critical_path, _, _ = self.find_critical_path()
            res = None

            # When all nodes are completed, return the output of the critical end node or end node.
            if end_nodes and len([node for node in end_nodes if node.status == TaskStatus.COMPLETED]) == len(end_nodes):
                if critical_end_node:
                    return critical_end_node.task.output, self.outputs
                else:
                    return [v.task.output for k, v in end_nodes.items()][0], self.outputs

            # Else, execute nodes connected with the critical_path
            elif critical_path:
                for item in critical_path:
                    edge = [v for k, v in self.edges.items() if item in k]
                    if edge:
                        edge = edge[0]

                        if edge.target.status == TaskStatus.COMPLETED:
                            res = edge.target.output

                        else:
                            res = edge.activate()
                            node_identifier = edge.target.identifier
                            self.outputs.update({ node_identifier: res })

                            if not res and start_nodes:
                                for node in start_nodes:
                                    res = node.handle_task_execution()
                                    self.outputs.update({ node.identifier: res })

            # If no critical paths in the graph, simply start from the start nodes.
            elif start_nodes:
                for node in start_nodes:
                    res = node.handle_task_execution()
                    self.outputs.update({ node.identifier: res })

            # If none of above is applicable, try to activate all the edges.
            else:
                for k, edge in self.edges.items():
                    res = edge.activate()
                    node_identifier = edge.target.identifier
                    self.outputs.update({ node_identifier: res })

            self.concl = res
            self.concl_template = self.concl_template if self.concl_template else res.pydantic.__class__ if res.pydantic else None
            # last_task_output = [v for v in self.outputs.values()][len([v for v in self.outputs.values()]) - 1] if [v for v in self.outputs.values()] else None
            return res, self.outputs



from abc import ABC, abstractmethod
from typing import Any, Dict, List, Optional

import contextlib
import hashlib
import io
import logging
import os
import shutil

import chromadb
import chromadb.errors
from chromadb.api import ClientAPI
from chromadb.api.types import OneOrMany
from chromadb.config import Settings

from versionhq._utils.logger import Logger
from versionhq._utils.vars import KNOWLEDGE_DIRECTORY
from versionhq.storage.utils import fetch_db_storage_path
from versionhq.knowledge.embedding import EmbeddingConfigurator


@contextlib.contextmanager
def suppress_logging(
    logger_name="chromadb.segment.impl.vector.local_persistent_hnsw",
    level=logging.ERROR,
):
    logger = logging.getLogger(logger_name)
    original_level = logger.getEffectiveLevel()
    logger.setLevel(level)
    with (
        contextlib.redirect_stdout(io.StringIO()),
        contextlib.redirect_stderr(io.StringIO()),
        contextlib.suppress(UserWarning),
    ):
        yield
    logger.setLevel(original_level)



class BaseKnowledgeStorage(ABC):
    """
    Abstract base class for knowledge storage implementations.
    """

    @abstractmethod
    def search(self, query: List[str], limit: int = 3, filter: Optional[dict] = None, score_threshold: float = 0.35) -> List[Dict[str, Any]]:
        """Search for documents in the knowledge base."""
        pass

    @abstractmethod
    def save(self, documents: List[str], metadata: Dict[str, Any] | List[Dict[str, Any]]) -> None:
        """Save documents to the knowledge base."""
        pass

    @abstractmethod
    def reset(self) -> None:
        """Reset the knowledge base."""
        pass



class KnowledgeStorage(BaseKnowledgeStorage):
    """
    A class to store ChromaDB Storage vals that handles embeddings, ChromaClient, and Collection.
    """

    app: Optional[ClientAPI] = None # store ChromaDBClient object
    collection_name: Optional[str] = "knowledge"
    collection: Optional[chromadb.Collection] = None
    embedding_function: Optional[Any] = None # store ChromaDB's EmbeddingFunction object
    embedder_config: Optional[Dict[str, Any]] = None # store config dict for embedding_function


    def __init__(self, embedder_config: Optional[Dict[str, Any]] = None, collection_name: Optional[str] = None):
        self.collection_name = self._validate_collection_name(collection_name) if collection_name else "knowledge"
        self.embedder_config = embedder_config
        self.initialize_knowledge_storage()


    def _validate_collection_name(self, collection_name: str = None) -> str:
        """
        Return a valid collection name from the given collection name.
        Expected collection name (1) contains 3-63 characters, (2) starts and ends with an alphanumeric character, (3) otherwise contains only alphanumeric characters, underscores or hyphens (-), (4) contains no two consecutive periods (..) and (5) is not a valid IPv4 address.
        """
        collection_name = collection_name if collection_name else self.collection_name
        valid_collection_name = collection_name.replace(' ', "-").replace("(", "-").replace(")", "").replace("..", "")
        return valid_collection_name


    def _create_default_embedding_function(self) -> Any:
        from chromadb.utils.embedding_functions.openai_embedding_function import OpenAIEmbeddingFunction

        return OpenAIEmbeddingFunction(
            api_key=os.getenv("OPENAI_API_KEY"), model_name="text-embedding-3-small"
        )


    def _set_embedding_function(self, embedder_config: Optional[Dict[str, Any]] = None) -> None:
        """
        Set the embedding configuration for the knowledge storage.
        """
        self.embedding_function = EmbeddingConfigurator().configure_embedder(embedder_config) if embedder_config else self._create_default_embedding_function()


    def initialize_knowledge_storage(self):
        """
        Create ChromaClinent, set up the embedding function using `embedder_config`, and get or create Collection.
        """
        base_path = os.path.join(fetch_db_storage_path(), "knowledge")
        chroma_client = chromadb.PersistentClient(path=base_path, settings=Settings(allow_reset=True))
        self.app = chroma_client
        self._set_embedding_function(embedder_config=self.embedder_config)
        self.collection_name = self.collection_name if self.collection_name else "knowledge"

        try:
            if self.app:
                self.collection = self.app.get_or_create_collection(name=self.collection_name, embedding_function=self.embedding_function)
            else:
                raise Exception("Vector Database Client not initialized")
        except Exception:
            raise Exception("Failed to create or get collection")


    def search(self, query: List[str], limit: int = 3, filter: Optional[dict] = None, score_threshold: float = 0.35) -> List[Dict[str, Any]]:
        with suppress_logging():
            if self.collection:
                fetched = self.collection.query(query_texts=query, n_results=limit, where=filter)
                results = []
                for i in range(len(fetched["ids"][0])):
                    result = {
                        "id": fetched["ids"][0][i],
                        "metadata": fetched["metadatas"][0][i],
                        "context": fetched["documents"][0][i],
                        "score": fetched["distances"][0][i],
                    }
                    if result["score"] >= score_threshold:
                        results.append(result)
                return results
            else:
                raise Exception("Collection not initialized")


    def save(self, documents: List[str], metadata: Optional[Dict[str, Any] | List[Dict[str, Any]]] = None) -> None:
        if not self.collection:
            self.initialize_knowledge_storage()
            # raise Exception("Collection not initialized")

        try:
            unique_docs = {}
            for i, doc in enumerate(documents):
                if doc:
                    doc = doc
                    if isinstance(doc, list):
                        doc = doc[0]

                    doc_id = hashlib.sha256(str(doc).encode("utf-8")).hexdigest()
                    doc_metadata = None
                    if metadata:
                        if isinstance(metadata, list):
                            doc_metadata = metadata[i]
                        else:
                            doc_metadata = metadata
                    unique_docs[doc_id] = (doc, doc_metadata)

            filtered_docs = []
            filtered_metadata = []
            filtered_ids = []

            for doc_id, (doc, meta) in unique_docs.items():
                if doc_id and doc:
                    filtered_docs.append(doc)
                    filtered_metadata.append(meta)
                    filtered_ids.append(doc_id)

            final_metadata: Optional[OneOrMany[chromadb.Metadata]] = (
                None if all(m is None for m in filtered_metadata) else filtered_metadata
            )

            if filtered_docs:
                self.collection.upsert(documents=filtered_docs, metadatas=final_metadata, ids=filtered_ids)

        except chromadb.errors.InvalidDimensionException as e:
            Logger(verbose=True).log(
                level="error",
                message="Embedding dimension mismatch. This usually happens when mixing different embedding models.",
                color="red",
            )
            raise ValueError("Embedding dimension mismatch. Make sure you're using the same embedding model across all operations with this collection.") from e

        except Exception as e:
            Logger(verbose=True).log(level="error", message=f"Failed to upsert documents: {str(e)}", color="red")
            raise


    def reset(self):
        base_path = os.path.join(fetch_db_storage_path(), KNOWLEDGE_DIRECTORY)
        if not self.app:
            self.app = chromadb.PersistentClient(path=base_path, settings=Settings(allow_reset=True))
        self.app.reset()
        shutil.rmtree(base_path)
        self.app = None
        self.collection = None



import json
import sqlite3
import datetime
from typing import Any, Dict, List, Optional

from versionhq._utils.logger import Logger
from versionhq.storage.utils import fetch_db_storage_path

storage_path = fetch_db_storage_path()
default_db_name = "task_output"


class TaskOutputSQLiteStorage:
    """
    An SQLite storage class to handle storing task outputs.
    """

    def __init__(self, db_path: str = f"{storage_path}/{default_db_name}.db") -> None:
        self.db_path = db_path
        self._logger = Logger(verbose=True)
        self._initialize_db()


    def _initialize_db(self):
        """
        Initializes the SQLite database and creates LTM table.
        """

        try:
            with sqlite3.connect(self.db_path) as conn:
                cursor = conn.cursor()
                cursor.execute(
                    """
                    CREATE TABLE IF NOT EXISTS task_output (
                        task_id TEXT PRIMARY KEY,
                        output JSON,
                        task_index INTEGER,
                        inputs JSON,
                        was_replayed BOOLEAN,
                        timestamp DATETIME DEFAULT CURRENT_TIMESTAMP
                    )
                """
                )
                conn.commit()

        except sqlite3.Error as e:
            self._logger.log(level="error", message=f"SQL database initialization failed: {str(e)}", color="red")


    def add(self, task, output: Dict[str, Any], task_index: int, was_replayed: bool = False, inputs: Dict[str, Any] = {}):
        try:
            with sqlite3.connect(self.db_path) as conn:
                cursor = conn.cursor()
                cursor.execute(
                        """INSERT OR REPLACE INTO task_output
                            (task_id, output, task_index, inputs, was_replayed, timestamp)
                            VALUES (?, ?, ?, ?, ?, ?)
                        """,
                     (str(task.id), json.dumps(output), task_index, json.dumps(inputs), was_replayed, datetime.datetime.now())
                )
                conn.commit()

        except sqlite3.Error as e:
            self._logger.log(level="error", message=f"SAVING TASK OUTPUTS ERROR: {e}", color="red")


    def update(self, task_index: int, **kwargs):
        try:
            with sqlite3.connect(self.db_path) as conn:
                cursor = conn.cursor()
                fields, values = [], []
                for k, v in kwargs.items():
                    fields.append(f"{k} = ?")
                    values.append(json.dumps(v) if isinstance(v, dict) else v)

                query = f"UPDATE latest_kickoff_task_output SET {', '.join(fields)} WHERE task_index = ?"
                values.append(task_index)
                cursor.execute(query, tuple(values))
                conn.commit()

                if cursor.rowcount == 0:
                    self._logger.log(
                        level="warning", message=f"No row found with task_index {task_index}. No update performed.", color="yellow",
                    )

        except sqlite3.Error as e:
            self._logger.log(level="error", message=f"UPDATE TASK OUTPUTS ERROR: {e}", color="red")


    def load(self) -> Optional[List[Dict[str, Any]]]:
        try:
            with sqlite3.connect(self.db_path) as conn:
                cursor = conn.cursor()
                cursor.execute("""
                SELECT *
                FROM task_output
                ORDER BY task_index
                """)

                rows = cursor.fetchall()
                results = []
                for row in rows:
                    result = {
                        "task_id": row[0],
                        "output": json.loads(row[1]),
                        "task_index": row[2],
                        "inputs": json.loads(row[3]),
                        "was_replayed": row[4],
                        "timestamp": row[5],
                    }
                    results.append(result)
                return results

        except sqlite3.Error as e:
            self._logger.log(level="error", message=f"LOADING TASK OUTPUTS ERROR: {e}", color="red")
            return None


    def delete_all(self):
        try:
            with sqlite3.connect(self.db_path) as conn:
                cursor = conn.cursor()
                cursor.execute("DELETE FROM task_output")
                conn.commit()

        except sqlite3.Error as e:
            self._logger.log(level="error", message=f"ERROR: Failed to delete all: {e}", color="red")



from typing import Type

JSON_URL = "https://raw.githubusercontent.com/BerriAI/litellm/main/model_prices_and_context_window.json"

PROVIDERS = [
    "openai",
    "gemini",
    "openrouter",
    "huggingface",
    "anthropic",
    "sagemaker",
    "bedrock",
    "ollama",
    "watson",
    "azure",
    "cerebras",
    "llama",
]

ENDPOINT_PROVIDERS = [
    "huggingface",
]

"""
List of models available on the framework.
Model names align with the LiteLLM's key names defined in the JSON URL.
Provider names align with the custom provider or model provider names.
-> model_key = custom_provider_name/model_name

Option
litellm.pick_cheapest_chat_models_from_llm_provider(custom_llm_provider: str, n=1)
"""

MODELS = {
    "openai": [
        "gpt-4",
        "gpt-4o",
        "gpt-4o-mini",
        "o1-mini",
        "o1-preview",
    ],
    "gemini": [
        "gemini/gemini-1.5-flash",
        "gemini/gemini-1.5-pro",
        "gemini/gemini-2.0-flash-exp",
    ],
    "anthropic": [
        "claude-3-5-sonnet-20241022",
        "claude-3-5-sonnet-20240620",
        "claude-3-haiku-2024030",
        "claude-3-opus-20240229",
        "claude-3-haiku-20240307",
    ],
    "openrouter": [
        "openrouter/deepseek/deepseek-r1",
        "openrouter/qwen/qwen-2.5-72b-instruct",
        "openrouter/google/gemini-2.0-flash-thinking-exp:free",
        "openrouter/google/gemini-2.0-flash-thinking-exp-1219:free",
        "openrouter/google/gemini-2.0-flash-001",
        "openrouter/meta-llama/llama-3.3-70b-instruct",
        "openrouter/mistralai/mistral-large-2411",
    ],
    "huggingface": [
        "huggingface/qwen/qwen2.5-VL-72B-Instruct",
    ],
    # "sagemaker": [
    #     "sagemaker/huggingface-text2text-flan-t5-base",
    #     "sagemaker/huggingface-llm-gemma-7b",
    #     "sagemaker/jumpstart-dft-meta-textgeneration-llama-2-13b",
    #     "sagemaker/jumpstart-dft-meta-textgeneration-llama-2-70b",
    #     "sagemaker/jumpstart-dft-meta-textgeneration-llama-3-8b",
    #     "sagemaker/jumpstart-dft-meta-textgeneration-llama-3-70b",
    #     "sagemaker/huggingface-llm-mistral-7b"
    # ], #https://docs.aws.amazon.com/sagemaker/latest/dg/jumpstart-foundation-models-latest.html
    "ollama": [
      "ollama/llama3.1",
      "ollama/mixtral",
      "ollama/mixtral-8x22B-Instruct-v0.1",
    ],
    # "watson": [
    #     "watsonx/meta-llama/llama-3-1-70b-instruct",
    #     "watsonx/meta-llama/llama-3-1-8b-instruct",
    #     "watsonx/meta-llama/llama-3-2-11b-vision-instruct",
    #     "watsonx/meta-llama/llama-3-2-1b-instruct",
    #     "watsonx/meta-llama/llama-3-2-90b-vision-instruct",
    #     "watsonx/meta-llama/llama-3-405b-instruct",
    #     "watsonx/mistral/mistral-large",
    #     "watsonx/ibm/granite-3-8b-instruct",
    # ],
    "bedrock": [
        "bedrock/anthropic.claude-3-5-sonnet-20240620-v1:0",
        "bedrock/anthropic.claude-3-sonnet-20240229-v1:0",
        "bedrock/anthropic.claude-3-haiku-20240307-v1:0",
        "bedrock/anthropic.claude-3-opus-20240229-v1:0",
        "bedrock/anthropic.claude-v2",
        "bedrock/anthropic.claude-instant-v1",
        "bedrock/meta.llama3-1-405b-instruct-v1:0",
        "bedrock/meta.llama3-1-70b-instruct-v1:0",
        "bedrock/meta.llama3-1-8b-instruct-v1:0",
        "bedrock/meta.llama3-70b-instruct-v1:0",
        "bedrock/meta.llama3-8b-instruct-v1:0",
        "bedrock/amazon.titan-text-lite-v1",
        "bedrock/amazon.titan-text-express-v1",
        "bedrock/cohere.command-text-v14",
        "bedrock/ai21.j2-mid-v1",
        "bedrock/ai21.j2-ultra-v1",
        "bedrock/ai21.jamba-instruct-v1:0",
        "bedrock/meta.llama2-13b-chat-v1",
        "bedrock/meta.llama2-70b-chat-v1",
        "bedrock/mistral.mistral-7b-instruct-v0:2",
        "bedrock/mistral.mixtral-8x7b-instruct-v0:1",
    ],
}



KEYS = {
    "openai": ["OPENAI_API_KEY"],
    "gemini": ["GEMINI_API_KEY"],
    "anthropic": ["ANTHROPIC_API_KEY"],
    "huggingface": ["HUGGINGFACE_API_KEY", ],
    "sagemaker": ["AWS_ACCESS_KEY_ID", "ADW_SECURET_ACCESS_KEY", "AWS_REGION_NAME"],
}



"""
Max input token size by the model.
"""
LLM_CONTEXT_WINDOW_SIZES = {
    "gpt-4": 8192,
    "gpt-4o": 128000,
    "gpt-4o-mini": 128000,
    "gpt-4-turbo": 128000,
    "o1-preview": 128000,
    "o1-mini": 128000,

    "gemini/gemini-1.5-flash": 1048576,
    "gemini/gemini-1.5-pro": 2097152,
    "gemini/gemini-2.0-flash-exp": 1048576,
    "gemini/gemini-gemma-2-9b-it": 8192,
    "gemini/gemini-gemma-2-27b-it": 8192,

    "claude-3-5-sonnet-20241022": 200000,
    "claude-3-5-sonnet-20240620": 200000,
    "claude-3-sonnet-20240229": 200000,
    "claude-3-opus-20240229": 200000,
    "claude-3-haiku-20240307": 200000,
    "claude-3-5-sonnet-2024102": 200000,

    "deepseek-chat": 128000,
    "deepseek/deepseek-reasoner": 8192,

    "gemma2-9b-it": 8192,
    "gemma-7b-it": 8192,
    "llama3-groq-70b-8192-tool-use-preview": 8192,
    "llama3-groq-8b-8192-tool-use-preview": 8192,
    "llama-3.1-70b-versatile": 131072,
    "llama-3.1-8b-instant": 131072,
    "llama-3.2-1b-preview": 8192,
    "llama-3.2-3b-preview": 8192,
    "llama-3.2-11b-text-preview": 8192,
    "llama-3.2-90b-text-preview": 8192,
    "llama3-70b-8192": 8192,
    "llama3-8b-8192": 8192,
    "mixtral-8x7b-32768": 32768,
}



"""
Params for litellm.completion().
"""

PARAMS = {
    "litellm": [
        "api_base",
        "api_version,",
        "num_retries",
        "context_window_fallback_dict",
        "fallbacks",
        "metadata",
    ],
    "common": [
        "model",
        "messages",
        "temperature",
        "top_p",
        "max_tokens",
        "stream",
        "tools",
        "tool_choice",
        "response_format",
        "n",
        "stop",
        "base_url",
        "api_key",
    ],
    "openai": [
        "timeout",
        "stream_options",
        "max_compl,etion_tokens",
        "modalities",
        "prediction",
        "audio",
        "presence_penalty",
        "frequency_penalty",
        "logit_bias",
        "user",
        "seed",
        "logprobs",
        "top_logprobs",
        "parallel_tool_calls",
        "extra_headers",
        "model_list"
    ],
    "gemini": [
        "topK",
    ]
}


class SchemaType:
    """
    A class to store/convert a LLM-valid schema type from the Python Type object.
    https://swagger.io/docs/specification/v3_0/data-models/data-types/
    https://cloud.google.com/vertex-ai/docs/reference/rest/v1/Schema#Type
    """

    def __init__(self, type: Type):
        self.type = type

    def convert(self) -> str:
        if self.type is None:
            return "string"

        if self.type is int:
            return "integer"
        elif self.type is float:
            return "number"
        elif self.type is str:
            return "string"
        elif self.type is dict:
            return "object"
        elif self.type is list:
            return "array"
        elif self.type is bool:
            return "boolean"
        else:
            return "string"



from versionhq.knowledge.model import Knowledge, KnowledgeStorage
from versionhq.knowledge.source import (
    CSVKnowledgeSource,
    ExcelKnowledgeSource,
    PDFKnowledgeSource,
    TextFileKnowledgeSource,
    JSONKnowledgeSource,
    StringKnowledgeSource
)
from versionhq.knowledge.source_docling import DoclingSource

__all__ = [
    "Knowledge",
    "KnowledgeStorage",
    "DoclingSource",
    "CSVKnowledgeSource",
    "ExcelKnowledgeSource",
    "PDFKnowledgeSource",
    "TextFileKnowledgeSource",
    "JSONKnowledgeSource",
    "StringKnowledgeSource"
]



import json
import os
import sys
import threading
import warnings
from dotenv import load_dotenv
import litellm
from litellm import JSONSchemaValidationError
from contextlib import contextmanager
from typing import Any, Dict, List, Optional
from typing_extensions import Self
from pydantic import BaseModel, Field, PrivateAttr, model_validator, ConfigDict

from versionhq.llm.llm_vars import LLM_CONTEXT_WINDOW_SIZES, MODELS, PARAMS, PROVIDERS, ENDPOINT_PROVIDERS
from versionhq.tool.model import Tool, ToolSet
from versionhq._utils.logger import Logger


load_dotenv(override=True)
LITELLM_API_KEY = os.environ.get("LITELLM_API_KEY")
LITELLM_API_BASE = os.environ.get("LITELLM_API_BASE")
DEFAULT_CONTEXT_WINDOW_SIZE = int(8192 * 0.75)
DEFAULT_MODEL_NAME = os.environ.get("DEFAULT_MODEL_NAME", "gpt-4o-mini")
DEFAULT_MODEL_PROVIDER_NAME = os.environ.get("DEFAULT_MODEL_PROVIDER_NAME", "openai")

class FilteredStream:
    def __init__(self, original_stream):
        self._original_stream = original_stream
        self._lock = threading.Lock()

    def write(self, s) -> int:
        with self._lock:
            if (
                "Give Feedback / Get Help: https://github.com/BerriAI/litellm/issues/new"
                in s
                or "LiteLLM.Info: If you need to debug this error, use `os.environ['LITELLM_LOG'] = 'DEBUG'`"
                in s
            ):
                return 0
            return self._original_stream.write(s)

    def flush(self):
        with self._lock:
            return self._original_stream.flush()


@contextmanager
def suppress_warnings():
    with warnings.catch_warnings():
        litellm.set_verbose = False
        warnings.filterwarnings(action="ignore")
        old_stdout = sys.stdout
        old_stderr = sys.stderr
        sys.stdout = FilteredStream(old_stdout)
        sys.stderr = FilteredStream(old_stderr)

        try:
            yield
        finally:
            sys.stdout = old_stdout
            sys.stderr = old_stderr


class LLM(BaseModel):
    """
    An LLM class to store params to send to the LLM. Use LiteLLM or custom providers for the endpoint.
    """

    _logger: Logger = PrivateAttr(default_factory=lambda: Logger(verbose=True))
    _init_model_name: str = PrivateAttr(default=None)
    _tokens: int = PrivateAttr(default=0) # accumulate total tokens used for the call
    model_config = ConfigDict(extra="allow")

    model: str = Field(default=None)
    provider: Optional[str] = Field(default=None, description="model provider")
    endpoint_provider: Optional[str] = Field(default=None, description="custom endpoint provider for pass through llm call. must need base_url")
    base_url: Optional[str] = Field(default=None, description="api base url for endpoint provider")
    api_key: Optional[str] = Field(default=None, description="api key to access the model")

    # optional params
    response_format: Optional[Any] = Field(default=None)
    timeout: Optional[float | int] = Field(default=None)
    max_tokens: Optional[int] = Field(default=None)
    max_completion_tokens: Optional[int] = Field(default=None)
    context_window_size: Optional[int] = Field(default=DEFAULT_CONTEXT_WINDOW_SIZE)
    temperature: Optional[float] = Field(default=None)
    top_p: Optional[float] = Field(default=None)
    n: Optional[int] = Field(default=None)
    stop: Optional[str | List[str]] = Field(default=None)
    presence_penalty: Optional[float] = Field(default=None)
    frequency_penalty: Optional[float] = Field(default=None)
    logit_bias: Optional[Dict[int, float]] = Field(default=None)
    seed: Optional[int] = Field(default=None)
    logprobs: Optional[bool] = Field(default=None)
    top_logprobs: Optional[int] = Field(default=None)
    tools: Optional[List[Dict[str, Any]]] = Field(default_factory=list, description="store a list of tool properties")
    callbacks: List[Any] = Field(default_factory=list)
    other_valid_config: Optional[Dict[str, Any]] = Field(default_factory=dict, description="store other valid values in dict to cascade to the model")

    # LiteLLM specific fields
    api_base: Optional[str] = Field(default=None, description="litellm specific field - api base of the model provider")
    api_version: Optional[str] = Field(default=None)
    num_retries: Optional[int] = Field(default=1)
    context_window_fallback_dict: Optional[Dict[str, Any]] = Field(default=None, description="A mapping of model to use if call fails due to context window error")
    fallbacks: Optional[List[Any]]= Field(default=None, description="A list of model names + params to be used, in case the initial call fails")
    metadata: Optional[Dict[str, Any]] = Field(default=None)

    litellm.drop_params = True
    litellm.set_verbose = True
    os.environ['LITELLM_LOG'] = 'DEBUG'


    @model_validator(mode="after")
    def validate_model_and_provider(self) -> Self:
        """
        Validate the given model, provider, interface provider.
        """

        self._init_model_name = self.model

        if self.model is None and self.provider is None:
            self.model = DEFAULT_MODEL_NAME
            self.provider = DEFAULT_MODEL_PROVIDER_NAME

        elif self.model is None and self.provider:
            if self.provider not in PROVIDERS:
                self._logger.log(level="warning", message=f"Invalid model provider is provided. We will assign a default model.", color="yellow")
                self.model = DEFAULT_MODEL_NAME
                self.provider = DEFAULT_MODEL_PROVIDER_NAME

            else:
                provider_model_list = MODELS.get(self.provider)
                if provider_model_list:
                    self.model = provider_model_list[0]
                    self.provider = self.provider
                else:
                    self._logger.log(level="warning", message=f"This provider has not models to be called. We will assign a default model.", color="yellow")
                    self.model = DEFAULT_MODEL_NAME
                    self.provider = DEFAULT_MODEL_PROVIDER_NAME

        elif self.model and self.provider is None:
            model_match = [
                item for item in [
                    [val for val in v if val == self.model][0] for k, v in MODELS.items() if [val for val in v if val == self.model]
                ] if item
            ]
            model_partial_match = [
                item for item in [
                    [val for val in v if val.find(self.model) != -1][0] for k, v in MODELS.items() if [val for val in v if val.find(self.model) != -1]
                ] if item
            ]
            provider_match = [k for k, v in MODELS.items() if k == self.model]

            if model_match:
                self.model = model_match[0]
                self.provider = [k for k, v in MODELS.items() if self.model in v][0]

            elif model_partial_match:
                self.model = model_partial_match[0]
                self.provider = [k for k, v in MODELS.items() if [item for item in v if item.find(self.model) != -1]][0]

            elif provider_match:
                provider = provider_match[0]
                if self.MODELS.get(provider):
                    self.provider = provider
                    self.model = self.MODELS.get(provider)[0]
                else:
                    self.provider = DEFAULT_MODEL_PROVIDER_NAME
                    self.model = DEFAULT_MODEL_NAME

            else:
                self.model = DEFAULT_MODEL_NAME
                self.provider = DEFAULT_MODEL_PROVIDER_NAME

        else:
            provider_model_list = MODELS.get(self.provider)
            if self.model not in provider_model_list:
                self._logger.log(level="warning", message=f"The provided model: {self._init_model_name} is not in the list. We will assign a default model.", color="yellow")
                self.model = DEFAULT_MODEL_NAME
                self.provider = DEFAULT_MODEL_PROVIDER_NAME

        # trigger pass-through custom endpoint.
        if self.provider in ENDPOINT_PROVIDERS:
            self.endpoint_provider = self.provider

        return self


    @model_validator(mode="after")
    def validate_model_params(self) -> Self:
        """
        Set up valid params to the model after setting up a valid model, provider, interface provider names.
        """
        self._tokens = 0

        if self.callbacks:
            self._set_callbacks(self.callbacks)

        self.context_window_size = self._get_context_window_size()

        api_key_name = self.provider.upper() + "_API_KEY" if self.provider else None
        if api_key_name:
            self.api_key = os.environ.get(api_key_name, None)

        base_url_key_name = self.endpoint_provider.upper() + "_API_BASE" if self.endpoint_provider else None

        if base_url_key_name:
            self.base_url = os.environ.get(base_url_key_name)
            self.api_base = self.base_url

        return self


    def _create_valid_params(self, config: Dict[str, Any]) -> Dict[str, Any]:
        """
        Return valid params (model + litellm original params) from the given config dict.
        """

        valid_params, valid_keys = dict(), list()

        if self.model:
            valid_keys = litellm.get_supported_openai_params(model=self.model, custom_llm_provider=self.endpoint_provider, request_type="chat_completion")

        if not valid_keys:
            valid_keys = PARAMS.get("common")

        valid_keys += PARAMS.get("litellm")

        for item in valid_keys:
            if hasattr(self, item) and getattr(self, item):
                valid_params[item] = getattr(self, item)
            elif item in self.other_valid_config and self.other_valid_config[item]:
                valid_params[item] = self.other_valid_config[item]
            elif item in config and config[item]:
                valid_params[item] = config[item]

        return valid_params


    def call(
        self,
        messages: List[Dict[str, str]],
        response_format: Optional[Dict[str, Any]] = None,
        tools: Optional[List[Tool | ToolSet | Any ]] = None,
        config: Optional[Dict[str, Any]] = {}, # any other conditions to pass on to the model.
        tool_res_as_final: bool = False
    ) -> str:
        """
        Execute LLM based on the agent's params and model params.
        """

        litellm.drop_params = True

        with suppress_warnings():
            if len(self.callbacks) > 0:
                self._set_callbacks(self.callbacks) # passed by agent

            try:
                res, tool_res = None, ""

                if not tools:
                    self.response_format = response_format
                    params = self._create_valid_params(config=config)
                    res = litellm.completion(model=self.model, messages=messages, stream=False, **params)
                    self._tokens += int(res["usage"]["total_tokens"])
                    return res["choices"][0]["message"]["content"]

                else:
                    try:
                        self.response_format = { "type": "json_object" }  if tool_res_as_final and self.provider != "gemini" else response_format
                        self.tools = [item.tool.properties if isinstance(item, ToolSet) else item.properties for item in tools]
                        params = self._create_valid_params(config=config)
                        res = litellm.completion(model=self.model, messages=messages, **params)
                        tool_calls = res.choices[0].message.tool_calls

                        if tool_calls:
                            for item in tool_calls:
                                func_name = item.function.name
                                func_args = item.function.arguments
                                if not isinstance(func_args, dict):
                                    try:
                                        func_args = json.loads(json.dumps(eval(str(func_args))))
                                    except:
                                        pass

                                # find a tool whose name is matched with the retrieved func_name
                                matches = []
                                for tool in tools:
                                    tool_name = tool.tool.name if isinstance(tool, ToolSet) else tool.name
                                    tool_func_name = tool.tool.func.__name__ if isinstance(tool, ToolSet) else tool.func.__name__
                                    if tool_name.replace(" ", "_") == func_name or tool_func_name == func_name or tool_name == "random_func":
                                        matches.append(tool)
                                    else:
                                        pass

                                if matches:
                                    tool_to_execute = matches[0]
                                    tool_instance = tool_to_execute.tool if isinstance(tool_to_execute, ToolSet) else tool_to_execute
                                    params = tool_to_execute.kwargs if isinstance(tool_to_execute, ToolSet) else func_args
                                    tool_res_to_add = tool_instance.run(params=params) if params else tool_instance.run()

                                    if tool_res_as_final:
                                        if tool_res_to_add not in tool_res:
                                            tool_res += str(tool_res_to_add)
                                    else:
                                        messages.append(res.choices[0].message)
                                        messages.append({ "role": "tool", "tool_call_id": item.id, "content": str(tool_res_to_add) })

                        else:
                            if tool_res_as_final and tools and not tool_res:
                                for item in tools:
                                    tool_res_to_add = item.tool.run(params=item.kwargs) if isinstance(item, ToolSet) else item.run()
                                    if tool_res_to_add not in tool_res:
                                        tool_res += str(tool_res_to_add)
                                    else:
                                        pass

                    except:
                        if tool_res_as_final and tools and not tool_res:
                            for item in tools:
                                tool_res_to_add = item.tool.run(params=item.kwargs) if isinstance(item, ToolSet) else item.run()
                                if tool_res_to_add not in tool_res:
                                    tool_res += str(tool_res_to_add)
                                else:
                                    pass
                        elif tools and not tool_res:
                                tool_res = res["choices"][0]["message"]["content"]
                        else:
                            pass


                if tool_res_as_final:
                    return tool_res
                else:
                    res = litellm.completion(model=self.model, messages=messages, **params)
                    self._tokens += int(res["usage"]["total_tokens"])
                    return res.choices[0].message.content


            except JSONSchemaValidationError as e:
                self._logger.log(level="error", message="Raw Response: {}".format(e.raw_response), color="red")
                raise e

            except Exception as e:
                self._logger.log(level="error", message=f"{self.model} failed to execute: {str(e)}", color="red")
                if "litellm.RateLimitError" in str(e):
                    raise e


    def _supports_function_calling(self) -> bool:
        try:
            if self.model:
                params = litellm.get_supported_openai_params(model=self.model)
                return "response_format" in params if params else False
        except Exception as e:
            self._logger.log(level="warning", message=f"Failed to get supported params: {str(e)}", color="yellow")
            return False


    def _supports_stop_words(self) -> bool:
        supported_params = litellm.get_supported_openai_params(model=self.model, custom_llm_provider=self.endpoint_provider)
        return "stop" in supported_params if supported_params else False


    def _get_context_window_size(self) -> int:
        """
        Only use 75% of the context window size to avoid cutting the message in the middle.
        """
        return int(LLM_CONTEXT_WINDOW_SIZES.get(self.model) * 0.75) if LLM_CONTEXT_WINDOW_SIZES.get(self.model) is not None else DEFAULT_CONTEXT_WINDOW_SIZE


    def _set_callbacks(self, callbacks: List[Any]):
        callback_types = [type(callback) for callback in callbacks]
        for callback in litellm.success_callback[:]:
            if type(callback) in callback_types:
                litellm.success_callback.remove(callback)

        for callback in litellm._async_success_callback[:]:
            if type(callback) in callback_types:
                litellm._async_success_callback.remove(callback)

        litellm.callbacks = callbacks



from typing import Any, Dict, Optional

from pydantic import BaseModel, PrivateAttr, Field, InstanceOf


class CacheHandler(BaseModel):
    """
    A class to add or read cache
    """

    _cache: Dict[str, Any] = PrivateAttr(default_factory=dict)

    def add(self, tool_name: str, input: str, output: Any) -> None:
        self._cache[f"{tool_name}-{input}"] = output

    def read(self, tool_name: str, input: str) -> Optional[str]:
        return self._cache.get(f"{tool_name}-{input}")



class CacheTool(BaseModel):
    """
    A cache tool to read the cached result.
    """

    name: str = "Cache Tool"
    cache_handler: InstanceOf[CacheHandler] = Field(default_factory=CacheHandler)

    def read_cache(self, key):
        split = key.split("tool:")
        tool = split[1].split("|input:")[0].strip()
        tool_input = split[1].split("|input:")[1].strip()
        return self.cache_handler.read(tool, tool_input)

    def tool(self):
        return Tool(
            func=self.read_cache,
            name=self.name,
            description="Read from cache"
        )




"""Test use cases on core/tool.md"""

def test_docs_core_tool_a():
    from versionhq import Tool

    def demo_func(message: str) -> str:
        return message + "_demo"

    my_tool = Tool(func=demo_func)
    res = my_tool.run(params=dict(message="Hi!"))

    assert res == "Hi!_demo"


def test_docs_core_tool_b():
    from versionhq import Tool, Agent

    def demo_func() -> str:
        return "demo"

    my_tool = Tool(func=demo_func)

    agent = Agent(
        role="Tool Handler",
        goal="efficiently use the given tools",
        tools=[my_tool, ]
    )

    assert agent.tools == [my_tool, ]
    assert agent.func_calling_llm.model is not None


def test_docs_core_tool_c():
    from versionhq import Tool, ToolSet, Agent

    def demo_func(message: str) -> str:
        return message + "_demo"

    tool_a = Tool(func=demo_func)
    toolset = ToolSet(tool=tool_a, kwargs={"message": "Hi"})
    agent = Agent(
        role="Tool Handler",
        goal="efficiently use the given tools",
        tools=[toolset]
    )

    assert agent.tools == [toolset]
    assert agent.func_calling_llm.model is not None


def test_docs_core_tool_d():
    from versionhq import Tool

    my_tool = Tool(name="my empty tool", func = lambda x: x)
    assert my_tool.name == "my empty tool" and my_tool.func is not None


    def demo_func() -> str:
        return "demo"

    my_tool_2 = Tool(func=demo_func)
    assert my_tool_2.name == "demo_func" and my_tool_2.func == demo_func


def test_docs_core_tool_e1():
    from versionhq import Tool

    my_tool = Tool(func=lambda x: f"demo-{x}" )
    res = my_tool._run(x="TESTING")
    assert res == "demo-TESTING"


def test_docs_core_tool_e2():
    from versionhq import Tool

    def demo_func() -> str:
        """...some complex execution..."""
        return "demo"

    my_tool = Tool(func=demo_func)
    res = my_tool._run()
    assert res == "demo"


def test_docs_core_tool_e3():
    from versionhq import Tool

    def demo_func(message: str) -> str:
        return message + "_demo"

    my_tool = Tool(func=demo_func)
    res = my_tool._run(message="Hi!")

    assert res == "Hi!_demo"


def test_docs_core_tool_f():
    from typing import Callable
    from versionhq import Tool

    class MyCustomTool(Tool):
        name: str = "custom tool"
        func: Callable

    my_custom_tool = MyCustomTool(func=lambda x: len(x))
    res = my_custom_tool._run(["demo1", "demo2"])
    assert res == 2


def test_docs_core_tool_g():
    from typing import List, Any, Callable
    from versionhq import Tool

    class CustomTool(Tool):
        name: str = "custom tool"
        func: Callable

    def demo_func(demo_list: List[Any]) -> int:
        return len(demo_list)

    my_tool = CustomTool(func=demo_func)
    res = my_tool.run(params=dict(demo_list=["demo1", "demo2"]))

    from versionhq import ToolHandler
    assert res == 2
    assert isinstance(my_tool.tool_handler, ToolHandler)


def test_docs_core_tool_h1():
    from versionhq import Tool, Agent, Task

    my_tool = Tool(name="demo tool", func=lambda x: "demo func")
    assert my_tool

    agent = Agent(
        role="demo",
        goal="execute tools",
        func_calling_llm="gpt-4o",
        tools=[my_tool],
        maxit=1,
        max_tokens=3000
    )
    assert agent.tools == [my_tool]

    task = Task(
        description="execute tools",
        can_use_agent_tools=True,
        tool_res_as_final=True
    )

    res = task.execute(agent=agent)
    assert res.raw == "demo func"


def test_docs_core_tool_h2():
    from versionhq import Tool, Agent, Task

    def demo_func(): return "demo func"
    my_tool = Tool(name="demo tool", func=demo_func)
    agent = Agent(
        role="demo",
        goal="execute the given tools",
        llm="gemini-2.0",
        tools=[my_tool],
        maxit=1,
        max_tokens=3000
    )

    task = Task(
        description="execute the given tools",
        can_use_agent_tools=True,
        tool_res_as_final=True
    )

    res = task.execute(agent=agent)
    assert res.raw == "demo func" and res.tool_output == "demo func"


def test_docs_core_tool_h3():
    from versionhq import Tool, Agent, Task

    my_tool = Tool(name="demo tool", func=lambda x: "demo func")

    agent = Agent(
        role="Demo Tool Handler",
        goal="execute tools",
        tools=[my_tool],
        maxit=1,
        max_tokens=3000
    )

    task = Task(
        description="execute tools",
        can_use_agent_tools=True,
        tool_res_as_final=True
    )

    res = task.execute(agent=agent)

    assert res.tool_output == "demo func"
    assert "Demo Tool Handler" in task.processed_agents


def test_docs_core_tool_h4():
    from versionhq.llm.model import LLM

    llm = LLM(model="gpt_4o")
    assert isinstance(llm._supports_function_calling(), bool)


def test_docs_core_tool_h5():
    from versionhq import Tool, ToolSet, Task, Agent

    def random_func(message: str) -> str:
        return message + "_demo"

    tool = Tool(name="tool", func=random_func)

    tool_set = ToolSet(
        tool=tool,
        kwargs=dict(message="empty func")
    )
    task = Task(
        description="execute the function",
        tools=[tool_set,], # use ToolSet to call args
        tool_res_as_final=True
    )

    res = task.execute()
    assert res.tool_output == "empty func_demo"


def test_docs_core_tool_i():
    from versionhq.tool.decorator import tool

    @tool("demo")
    def my_tool(test_words: str) -> str:
        """Test a tool decorator."""
        return test_words

    assert my_tool.name == "demo"
    assert "Tool: demo" in my_tool.description and "'test_words': {'description': '', 'type': 'str'" in my_tool.description
    assert my_tool.func("testing") == "testing"



import uuid
from abc import ABC, abstractmethod
from typing import Any, Dict, List, Callable, Type, Optional, get_args, get_origin
from pydantic import UUID4, InstanceOf, BaseModel, ConfigDict, Field, create_model, field_validator, model_validator
from pydantic_core import PydanticCustomError

from versionhq.clients.product.model import Product, ProductProvider
from versionhq.clients.customer import Status


class BaseCustomer(ABC, BaseModel):
    """
    Abstract base class for the base customer storing current status on the workflow and deployment method.
    """
    status: Status = Field(default=Status.NOT_ASSIGNED)

    @abstractmethod
    def _deploy(self, *args, **kwargs) -> Any:
        """Any method to deploy targeting the customer"""


    def deploy(self, *args, **kwargs) -> Any:
        if self.status is Status.READY_TO_DEPLOY:
            return self._deploy(self, **args, **kwargs)



class Customer(BaseCustomer):
    """
    Customer class to store customer info and handle deployment methods.
    """

    id: UUID4 = Field(default_factory=uuid.uuid4, frozen=True)
    name: Optional[str] = Field(default=None, description="customer's name if any")
    products: Optional[List[Product]] = Field(default=list, description="store products that the customer is associated with")
    analysis: str = Field(default=None, description="store the latest analysis results on the customer")
    function: Optional[Callable] = Field(default=None, descripition="store deploy function")
    config: Optional[Dict[str, Any]] = Field(default=None, description="config to the function")


    @field_validator("id", mode="before")
    @classmethod
    def _deny_user_set_id(cls, v: Optional[UUID4]) -> None:
        if v:
            raise PydanticCustomError("may_not_set_field", "This field is not to be set by the user.", {})


    def fetch_product_providers(self) -> List[ProductProvider] | None:
        """
        Return list of ProductProvider if the customer has `product_list`
        """
        res = []
        if self.products:
            for item in self.products:
                if item.provider not in res:
                    res.appned(item.provider)
        return res

    def _deploy(self, *args, **kwargs):
        return self.deploy(self, *args, **kwargs)


    def deploy(self, *args, **kwargs):
        self.status = Status.ACTIVE_ON_WORKFLOW

        if self.function:
            return self.function(**self.config)

        return super().deploy(*args, **kwargs)



import datetime
from typing import Any, Dict, List, Optional

from versionhq.storage.rag_storage import RAGStorage
from versionhq.storage.ltm_sqlite_storage import LTMSQLiteStorage
from versionhq._utils.logger import Logger


class MemoryData:
    """
    A class to store structured data to store in the memory.
    """
    def __init__(
        self,
        agent: Optional[str] = None, # task execution agent (core)
        task_description: Optional[str] = None,
        task_output: Optional[str] = None,
        config: Optional[Dict[str, Any]] = None
    ):
        self.agent = agent
        self.task_description = task_description
        self.task_output = task_output

        if config:
            for k, v in config.items():
                setattr(self, k, str(v))



class MemoryMetadata:
    """
     A class to store structured metadata to store in the memory.
    """

    def __init__(
        self,
        eval_criteria: Optional[str] = None,
        score: Optional[int | float] = None,
        suggestion: Optional[str] = None,
        eval_by: Optional[str] = None, # task evaluator agent
        config: Optional[Dict[str, Any]] = None
    ):
        self.eval_criteria = eval_criteria
        self.score = score
        self.suggestion = suggestion
        self.eval_by = eval_by

        if config:
            for k, v in config.items():
                setattr(self, k, str(v))


class Memory:
    """
    Base class for memory.
    """

    def __init__(self, storage: RAGStorage):
        self.storage = storage


    def save(
            self,
            data: MemoryData | Dict[str, Any],
            metadata: Optional[MemoryMetadata | Dict[str, Any]] = None,
            agent: Optional[str] = None
        ) -> None:

        """
        Create a dict for data and metadata without empty values before storing them in the given storage.
        """

        if not data:
            Logger(verbose=True).log(level="error", message="Missing data to store. Add either dict or MemoryData object", color="red")
            return None

        metadata_dict = metadata if isinstance(metadata, dict) else metadata.__dict__ if isinstance(metadata, MemoryMetadata) else dict()
        metadata_dict = {k: v for k, v in metadata_dict.items() if v} # remove empty values
        data_dict = data if isinstance(data, dict) else data.__dict__ if isinstance(data, MemoryData) else dict()
        data_dict = {k: v for k, v in data_dict.items() if v}

        if agent and data_dict["agent"] is None:
            data_dict["agent"] = agent

        if metadata_dict:
            self.storage.save(data=data_dict, metadata=metadata_dict)
        else:
            self.storage.save(data=data_dict)


    def search(self, query: str, limit: int = 3, score_threshold: float = 0.35) -> List[Any]:
        return self.storage.search(query=query, limit=limit, score_threshold=score_threshold)


class MemoryItem:
    """
    A class to store item to be saved in either long term memory or short term memory.
    """

    def __init__(self, data: MemoryData = None, metadata: Optional[MemoryMetadata] = None):
        self.data = data
        self.metadata = metadata if metadata is not None else {}


class ShortTermMemory(Memory):
    """
    A class for managing transient data related to immediate tasks and interactions.
    - Type: stm
    - Storage: Mem0Storage | RAGStorage
    """

    def __init__(self, agent = None, embedder_config: Dict[str, Any] = None, storage=None, path=None):
        if hasattr(agent, "memory_config") and agent.memory_config is not None:
            self.memory_provider = agent.memory_config.get("provider")
        else:
            self.memory_provider = None

        if self.memory_provider == "mem0":
            try:
                from versionhq.storage.mem0_storage import Mem0Storage
            except ImportError:
                try:
                    import os
                    os.system("uv add mem0ai --optional mem0ai")

                    from versionhq.storage.mem0_storage import Mem0Storage
                except:
                    raise ImportError("Mem0 is not installed. Please install it with `$ uv add versionhq[mem0ai]`.")

            storage = Mem0Storage(type="stm", agent=agent)
        else:
            storage = storage if storage else RAGStorage(type="stm", embedder_config=embedder_config, agents=[agent,], path=path)

        super().__init__(storage)


    def save(
            self,
            task_description: str = None,
            task_output: str = None,
            agent: Optional[str] = None,
            data: Optional[MemoryData] = None,
            metadata: Optional[MemoryMetadata] = None
        ) -> None:

        data = data if data else MemoryData(task_description=task_description, task_output=task_output, agent=agent)
        item = MemoryItem(data=data, metadata=metadata)

        if self.memory_provider == "mem0":
            item.data.task_output = f"Remember the following insights from Agent run: {item.data.task_output}"

        super().save(data=item.data.__dict__, metadata=item.metadata.__dict__ if item.metadata else {})


    def search(self, query: str, limit: int = 3, score_threshold: float = 0.35,):
        return self.storage.search(query=query, limit=limit, score_threshold=score_threshold)


    def reset(self) -> None:
        try:
            self.storage.reset()
        except Exception as e:
            raise Exception(f"An error occurred while resetting the short-term memory: {str(e)}")


class LongTermMemory(Memory):
    """
    A class for managing cross runs data related to overall task executions.
    - Type: ltm
    - Storage: LTMSQLiteStorage | RAGStorage
    """

    def __init__(self, storage=None, path=None):
        if not storage:
            storage = LTMSQLiteStorage(db_path=path) if path else LTMSQLiteStorage()

        super().__init__(storage)


    def save(
            self,
            task_description: str = None,
            task_output: str = None,
            agent: Optional[str] = None,
            data: Optional[MemoryData] = None,
            metadata: Optional[MemoryMetadata] = None
        ) -> None:

        data = data if data else MemoryData(task_description=task_description, task_output=task_output, agent=agent)
        item = MemoryItem(data=data, metadata=metadata)
        super().save(data=item.data, metadata=item.metadata)


    def search(self, query: str, latest_n: int = 3) -> List[Dict[str, Any]]:
        """
        Query the storage and return the results up to latest_n.
        """
        return self.storage.load(query=query, latest_n=latest_n)


    def reset(self) -> None:
        self.storage.reset()



class UserMemoryItem:
    def __init__(self, data: Any, user: str, metadata: Optional[Dict[str, Any]] = None):
        self.data = data
        self.user = user
        self.metadata = metadata if metadata is not None else {} # can be stored last purchased item, comm related to the user


class UserMemory(Memory):
    """
    UserMemory class for handling user memory storage and retrieval.
    - Type: user
    - Storage: Mem0Storage
    - Requirements: `user_id` in metadata
    """

    def __init__(self, agent=None, user_id=None):
        try:
            from versionhq.storage.mem0_storage import Mem0Storage
        except ImportError:
            raise ImportError("Mem0 is not installed. Please install it with `uv pip install mem0ai`.")

        if not user_id:
            raise ValueError("Need User Id to create UserMemory.")

        else:
            storage = Mem0Storage(type="user", agent=agent, user_id=user_id)
            super().__init__(storage)


    def save(self, value: str, metadata: Optional[Dict[str, Any]] = None, agent: Optional[str] = None) -> None:
        data = f"Remember the details about the user: {value}"
        super().save(value=data, metadata=metadata, agent=agent)


    def search(self, query: str, limit: int = 3, score_threshold: float = 0.35):
        results = self.storage.search(query=query, limit=limit, score_threshold=score_threshold)
        return results



import re
from typing import Any
from json_repair import repair_json

from versionhq._utils.i18n import I18N


FINAL_ANSWER_ACTION = "Final Answer:"
MISSING_ACTION_AFTER_THOUGHT_ERROR_MESSAGE = "I did it wrong. Invalid Format: I missed the 'Action:' after 'Thought:'. I will do right next, and don't use a tool I have already used.\n"
MISSING_ACTION_INPUT_AFTER_ACTION_ERROR_MESSAGE = "I did it wrong. Invalid Format: I missed the 'Action Input:' after 'Action:'. I will do right next, and don't use a tool I have already used.\n"
FINAL_ANSWER_AND_PARSABLE_ACTION_ERROR_MESSAGE = "I did it wrong. Tried to both perform Action and give a Final Answer at the same time, I must do one or the other"


class AgentAction:
    thought: str
    tool: str
    tool_input: str
    text: str
    result: str

    def __init__(self, thought: str, tool: str, tool_input: str, text: str):
        self.thought = thought
        self.tool = tool
        self.tool_input = tool_input
        self.text = text


class AgentFinish:
    thought: str
    output: str
    text: str

    def __init__(self, thought: str, output: str, text: str):
        self.thought = thought
        self.output = output
        self.text = text


class OutputParserException(Exception):
    error: str

    def __init__(self, error: str):
        self.error = error


class AgentParser:

    agent: Any = None

    def __init__(self, agent: Any):
        self.agent = agent

    def parse(self, text: str) -> AgentAction | AgentFinish:
        thought = self._extract_thought(text)
        includes_answer = FINAL_ANSWER_ACTION in text
        regex = r"Action\s*\d*\s*:[\s]*(.*?)[\s]*Action\s*\d*\s*Input\s*\d*\s*:[\s]*(.*)"
        action_match = re.search(regex, text, re.DOTALL)
        if action_match:
            if includes_answer:
                raise OutputParserException(
                    f"{FINAL_ANSWER_AND_PARSABLE_ACTION_ERROR_MESSAGE}"
                )
            action = action_match.group(1)
            clean_action = self._clean_action(action)

            action_input = action_match.group(2).strip()

            tool_input = action_input.strip(" ").strip('"')
            safe_tool_input = self._safe_repair_json(tool_input)

            return AgentAction(thought, clean_action, safe_tool_input, text)

        elif includes_answer:
            final_answer = text.split(FINAL_ANSWER_ACTION)[-1].strip()
            return AgentFinish(thought, final_answer, text)

        if not re.search(r"Action\s*\d*\s*:[\s]*(.*?)", text, re.DOTALL):
            # self.agent.increment_formatting_errors()
            raise OutputParserException(
                f"{MISSING_ACTION_AFTER_THOUGHT_ERROR_MESSAGE}",
            )
        elif not re.search(
            r"[\s]*Action\s*\d*\s*Input\s*\d*\s*:[\s]*(.*)", text, re.DOTALL
        ):
            # self.agent.increment_formatting_errors()
            raise OutputParserException(MISSING_ACTION_INPUT_AFTER_ACTION_ERROR_MESSAGE)
        else:
            _i18n = I18N()
            format = self._i18n.slice("format_without_tools")
            error = f"{format}"
            # self.agent.increment_formatting_errors()
            raise OutputParserException(error)


    def _extract_thought(self, text: str) -> str:
        regex = r"(.*?)(?:\n\nAction|\n\nFinal Answer)"
        thought_match = re.search(regex, text, re.DOTALL)
        if thought_match:
            return thought_match.group(1).strip()
        return ""


    def _clean_action(self, text: str) -> str:
        """Clean action string by removing non-essential formatting characters."""
        return re.sub(r"^\s*\*+\s*|\s*\*+\s*$", "", text).strip()


    def _safe_repair_json(self, tool_input: str) -> str:
        UNABLE_TO_REPAIR_JSON_RESULTS = ['""', "{}"]

        if tool_input.startswith("[") and tool_input.endswith("]"):
            return tool_input

        tool_input = tool_input.replace('"""', '"')

        result = repair_json(tool_input)
        if result in UNABLE_TO_REPAIR_JSON_RESULTS:
            return tool_input

        return str(result)



from datetime import datetime
from typing import Any, Dict, List, Optional

from pydantic import BaseModel, Field

from versionhq.storage.task_output_storage import TaskOutputSQLiteStorage


class ExecutionLog(BaseModel):
    task_id: str
    output: Dict[str, Any]
    timestamp: datetime = Field(default_factory=datetime.now)
    task_index: int
    inputs: Dict[str, Any] = Field(default_factory=dict)
    was_replayed: bool = False

    def __getitem__(self, key: str) -> Any:
        return getattr(self, key)



class TaskOutputStorageHandler:

    def __init__(self):
        self.storage = TaskOutputSQLiteStorage()


    def update(self, task, task_index: int, was_replayed: bool = False, inputs: Dict[str, Any] = {}) -> None:
        """
        task: task instance
        """
        saved_outputs = self.load()
        if saved_outputs is None:
            raise ValueError("Logs cannot be None")

        self.add(task, task_index, was_replayed, inputs)


    def add(self, task, task_index: int, was_replayed: bool = False, inputs: Dict[str, Any] = {}) -> None:
        from versionhq.task.model import Task

        output_to_store = dict()

        if isinstance(task, Task):
            output_to_store = dict(
                description=str(task.description),
                raw=str(task.output.raw),
                responsible_agent=str(task.processed_agents),
            )

        self.storage.add(task=task, output=output_to_store, task_index=task_index, was_replayed=was_replayed, inputs=inputs)


    def reset(self) -> None:
        self.storage.delete_all()


    def load(self) -> Optional[List[Dict[str, Any]]]:
        return self.storage.load()



white = "#fafafa"
black = "#191b1b"
darkgrey = "#363636"
grey = "#414141"
primary = "#B9005B"
orange = "#ffb713"
lightgreen = "#b8fff5"
green = "#00d1b2"
darkgreen = "#169d87"
darkergreen = "#026b5d"


# $black: #191b1b;
# $true-white: #ffffff;
# $white: #fafafa;
# $grey: #414141;
# $light-grey: #f9f9f9;
# $border-color: #dedede;
# $primary: #B9005B;
# $primary-hovered: #9D034F;
# $gradient-main: linear-gradient(to right $black $primary);
# $linkedin-blue: #0077B5;
# $bulma-primary: #00d1b2;
# $bluma-primary-dark: #169d87;
# $bulma-grey-light: hsl(0 0%, 71%);
# $bulma-grey: hsl(0, 0%, 48%);
# $bulma-dark: hsl(0, 0%, 21%); //#363636
# $bulma-white-ter: #F5F5F5;



import contextlib
import io
import logging
import os
import shutil
import uuid
from abc import ABC, abstractmethod
from typing import Any, Dict, List, Optional

from chromadb.api import ClientAPI

from versionhq.knowledge.embedding import EmbeddingConfigurator
from versionhq._utils.vars import MAX_FILE_NAME_LENGTH
from versionhq.storage.utils import fetch_db_storage_path


@contextlib.contextmanager
def suppress_logging(
    logger_name="chromadb.segment.impl.vector.local_persistent_hnsw",
    level=logging.ERROR,
):
    logger = logging.getLogger(logger_name)
    original_level = logger.getEffectiveLevel()
    logger.setLevel(level)
    with (
        contextlib.redirect_stdout(io.StringIO()),
        contextlib.redirect_stderr(io.StringIO()),
        contextlib.suppress(UserWarning),
    ):
        yield
    logger.setLevel(original_level)


class BaseRAGStorage(ABC):
    """
    Base class for RAG-based Storage implementations.
    """

    app: Any | None = None

    def __init__(
        self,
        type: str,
        allow_reset: bool = True,
        embedder_config: Optional[Any] = None,
        agents: List[Any] = None,
    ):
        self.type = type
        self.allow_reset = allow_reset
        self.embedder_config = embedder_config
        self.agents = agents

    def _initialize_agents(self) -> str:
        if self.agents:
            return "_".join(
                [self._sanitize_role(agent.role) for agent in self.agents]
            )
        return ""

    @abstractmethod
    def _sanitize_role(self, role: str) -> str:
        """Sanitizes agent roles to ensure valid directory names."""
        pass

    @abstractmethod
    def save(self, value: Any, metadata: Dict[str, Any]) -> None:
        """Save a value with metadata to the storage."""
        pass

    @abstractmethod
    def search(self, query: str, limit: int = 3, filter: Optional[dict] = None, score_threshold: float = 0.35) -> List[Any]:
        """Search for entries in the storage."""
        pass

    @abstractmethod
    def reset(self) -> None:
        """Reset the storage."""
        pass

    @abstractmethod
    def _generate_embedding(self, text: str, metadata: Optional[Dict[str, Any]] = None) -> Any:
        """Generate an embedding for the given text and metadata."""
        pass

    @abstractmethod
    def _initialize_app(self):
        """Initialize the vector db."""
        pass

    def setup_config(self, config: Dict[str, Any]):
        """Setup the config of the storage."""
        pass

    def initialize_client(self):
        """Initialize the client of the storage. This should setup the app and the db collection"""
        pass



class RAGStorage(BaseRAGStorage):
    """
    Extends Storage to handle embeddings for memory entries, improving
    search efficiency.
    """

    app: ClientAPI | None = None

    def __init__(self, type, allow_reset=True, embedder_config=None, agents=list(), path=None):
        super().__init__(type, allow_reset, embedder_config, agents)
        agents = agents
        agents = [self._sanitize_role(agent.role) for agent in agents]
        agents = "_".join(agents)

        self.agents = agents
        self.storage_file_name = self._build_storage_file_name(type, agents)
        self.type = type
        self.allow_reset = allow_reset
        self.path = path
        self._initialize_app()


    def _set_embedder_config(self):
        configurator = EmbeddingConfigurator()
        self.embedder_config = configurator.configure_embedder(self.embedder_config)


    def _initialize_app(self) -> None:
        import chromadb
        from chromadb.config import Settings

        self._set_embedder_config()
        try:
            chroma_client = chromadb.PersistentClient(
                path=self.path if self.path else self.storage_file_name,
                settings=Settings(allow_reset=self.allow_reset),
            )
            self.app = chroma_client
            self.collection = self.app.get_collection(name=self.type, embedding_function=self.embedder_config)
        except Exception:
            if self.app:
                self.collection = self.app.create_collection(name=self.type, embedding_function=self.embedder_config)


    def _sanitize_role(self, role: str) -> str:
        """
        Sanitizes agent roles to ensure valid directory names.
        """
        return role.replace("\n", "").replace(" ", "_").replace("/", "_")


    def _build_storage_file_name(self, type: str, file_name: str) -> str:
        """
        Ensures file name does not exceed max allowed by OS
        """
        base_path = f"{fetch_db_storage_path()}/{type}"

        if len(file_name) > MAX_FILE_NAME_LENGTH:
            logging.warning(
                f"Trimming file name from {len(file_name)} to {MAX_FILE_NAME_LENGTH} characters."
            )
            file_name = file_name[:MAX_FILE_NAME_LENGTH]

        return f"{base_path}/{file_name}"


    def _create_default_embedding_function(self):
        from chromadb.utils.embedding_functions.openai_embedding_function import (
            OpenAIEmbeddingFunction,
        )

        return OpenAIEmbeddingFunction(
            api_key=os.getenv("OPENAI_API_KEY"), model_name="text-embedding-3-small"
        )


    def _generate_embedding(self, text: str, metadata: Optional[Dict[str, Any]]) -> None:
        if not hasattr(self, "app") or not hasattr(self, "collection"):
            self._initialize_app()

        if metadata:
            self.collection.add(documents=[text], metadatas=[metadata, ], ids=[str(uuid.uuid4())])

        else:
            self.collection.add(documents=[text], ids=[str(uuid.uuid4())])


    def save(self, data: Dict[str, Any] | str, metadata: Optional[Dict[str, Any]] = dict()) -> None:
        if not hasattr(self, "app") or not hasattr(self, "collection"):
            self._initialize_app()
        try:
            self._generate_embedding(text=str(data), metadata=metadata)
        except Exception as e:
            logging.error(f"Error during {self.type} save: {str(e)}")


    def search(self, query: str, limit: int = 3, filter: Optional[dict] = None, score_threshold: float = 0.35) -> List[Any]:
        if not hasattr(self, "app"):
            self._initialize_app()

        try:
            with suppress_logging():
                response = self.collection.query(query_texts=query, n_results=limit)

            results = []
            for i in range(len(response["ids"][0])):
                result = {
                    "id": response["ids"][0][i],
                    "metadata": response["metadatas"][0][i],
                    "context": response["documents"][0][i],
                    "score": response["distances"][0][i],
                }
                if result["score"] >= score_threshold:
                    results.append(result)

            return results
        except Exception as e:
            logging.error(f"Error during {self.type} search: {str(e)}")
            return []


    def reset(self) -> None:
        try:
            if self.app:
                self.app.reset()
                shutil.rmtree(f"{fetch_db_storage_path()}/{self.type}")
                self.app = None
                self.collection = None
        except Exception as e:
            if "attempt to write a readonly database" in str(e):
                pass
            else:
                raise Exception(f"An error occurred while resetting the {self.type} memory: {e}")



import os
from unittest.mock import patch
from typing import Callable, Any

from versionhq.agent.model import Agent
from versionhq.agent.TEMPLATES.Backstory import BACKSTORY_SHORT, BACKSTORY_FULL
from versionhq.llm.model import DEFAULT_MODEL_NAME, LLM
from versionhq.tool.model import Tool
from versionhq.tool.decorator import tool

MODEL_NAME = os.environ.get("DEFAULT_MODEL_NAME", "gpt-3.5-turbo")
LITELLM_API_KEY = os.environ.get("LITELLM_API_KEY")


def test_build_agent_with_minimal_input():
    agent = Agent(
        role="analyst",
        goal="analyze the company's website and retrieve the product overview"
    )

    assert agent.role == "analyst"
    assert agent.backstory == BACKSTORY_SHORT.format(role=agent.role.lower(), goal=agent.goal.lower())
    assert isinstance(agent.llm, LLM)
    assert agent.llm.model == DEFAULT_MODEL_NAME
    assert agent.llm.api_key == LITELLM_API_KEY
    assert agent.tools == []
    assert isinstance(agent.func_calling_llm, LLM)


def test_build_agent_from_config():
    agent = Agent(config=dict(role="analyst", goal="analyze the company's website and retrieve the product overview"))

    assert agent.role == "analyst"
    assert agent.backstory == BACKSTORY_SHORT.format(role=agent.role.lower(), goal=agent.goal.lower())
    assert isinstance(agent.llm, LLM)
    assert agent.llm.model == DEFAULT_MODEL_NAME
    assert agent.llm.api_key == LITELLM_API_KEY
    assert agent.tools == []
    assert isinstance(agent.func_calling_llm, LLM)


def test_build_agent_with_backstory():
    agent = Agent(
        role="analyst",
        goal="analyze the company's website and retrieve the product overview",
        backstory="You are competitive analysts who have abundand knowledge in marketing, product management."
    )

    assert agent.role == "analyst"
    assert agent.goal == "analyze the company's website and retrieve the product overview"
    assert agent.backstory == "You are competitive analysts who have abundand knowledge in marketing, product management."
    assert isinstance(agent.llm, LLM)
    assert agent.llm.model == DEFAULT_MODEL_NAME
    assert agent.llm.api_key == LITELLM_API_KEY
    assert agent.tools == []
    assert isinstance(agent.func_calling_llm, LLM)


def test_build_agent():
    agent = Agent(
        role="analyst",
        goal="analyze the company's website and retrieve the product overview",
        skillsets=["financial analysis", "product management", ]
    )

    assert agent.role == "analyst"
    assert agent.backstory == BACKSTORY_FULL.format(role=agent.role.lower(), goal=agent.goal.lower(), skills=", ".join([item for item in agent.skillsets]), tools="")
    assert isinstance(agent.llm, LLM)
    assert agent.llm.model == DEFAULT_MODEL_NAME
    assert agent.llm.api_key == LITELLM_API_KEY
    assert agent.tools == []
    assert isinstance(agent.func_calling_llm, LLM)


def test_build_agent_with_llm():
    agent = Agent(
        role="analyst",
        goal="analyze the company's website and retrieve the product overview",
        skillsets=["financial analysis", "product management", ],
        llm="gpt-4o"
    )

    assert agent.role == "analyst"
    assert agent.role in agent.backstory
    assert agent.goal in agent.backstory
    assert [item in agent.backstory for item in agent.skillsets]
    assert isinstance(agent.llm, LLM)
    assert agent.llm.model == "gpt-4o"
    assert agent.llm.api_key == LITELLM_API_KEY
    assert agent.tools == []
    assert isinstance(agent.func_calling_llm, LLM)


def test_build_agent_with_llm_config():
    def dummy_func() -> str:
        return "dummy"

    llm_params = dict(deployment_name="gemini-1.5", max_tokens=4000, logprobs=False, abc="dummy key")
    llm_config = dict(
        temperature=1,
        top_p=0.1,
        n=1,
        stream=False,
        stream_options=None,
        stop="test",
        max_completion_tokens=10000,
        dummy="I am dummy"
    )
    agent = Agent(
        role="analyst",
        goal="run test on llm instance",
        llm=llm_params,
        callbacks=[dummy_func],
        llm_config=llm_config
    )

    assert isinstance(agent.llm, LLM)
    assert isinstance(agent.func_calling_llm, LLM)
    assert agent.llm.model == "gemini/gemini-1.5-flash"
    assert agent.llm.callbacks == [dummy_func]

    import litellm
    valid_params = litellm.get_supported_openai_params(model="gemini/gemini-1.5-flash")
    config = llm_params.update(llm_config)
    for key in valid_params:
        if config and [k == key for k, v in config.items()]:
            assert getattr(agent.llm, key) == config[key]


def test_build_agent_with_llm_instance():
    def dummy_func() -> str:
        return "dummy"

    llm = LLM(model="gemini-1.5", max_tokens=4000, logprobs=False)
    agent = Agent(
        role="analyst",
        goal="analyze the company's website and retrieve the product overview",
        llm=llm,
        callbacks=[dummy_func],
    )
    assert isinstance(agent.llm, LLM)
    assert agent.llm.model == "gemini/gemini-1.5-flash"
    assert agent.llm.api_key is not None
    assert agent.llm.max_tokens == 4000
    assert agent.llm.logprobs == False
    assert agent.llm.callbacks == [dummy_func]
    assert isinstance(agent.func_calling_llm, LLM)


def test_build_agent_with_llm_and_func_llm_config():
    def dummy_func() -> str:
        return "dummy"

    llm_params = dict(deployment_name="gemini-1.5", max_tokens=4000, logprobs=False, abc="dummy key")
    agent = Agent(
        role="analyst",
        goal="analyze the company's website and retrieve the product overview",
        func_calling_llm=llm_params,
        callbacks=[dummy_func]
    )

    assert isinstance(agent.llm, LLM) and agent.llm.model == DEFAULT_MODEL_NAME
    assert isinstance(agent.func_calling_llm, LLM)
    assert agent.func_calling_llm.model == "gemini/gemini-1.5-flash" if agent.func_calling_llm._supports_function_calling() else DEFAULT_MODEL_NAME


def test_build_agent_with_llm_and_func_llm_instance():
    def dummy_func() -> str:
        return "dummy"

    llm = LLM(model="gemini-1.5", max_tokens=4000, logprobs=False)
    agent = Agent(
        role="analyst",
        goal="analyze the company's website and retrieve the product overview",
        llm=llm,
        func_calling_llm=llm,
        llm_config=dict(),
        callbacks=[dummy_func]
    )
    assert isinstance(agent.llm, LLM) and isinstance(agent.func_calling_llm, LLM)
    assert agent.func_calling_llm.model == "gemini/gemini-1.5-flash" if agent.func_calling_llm._supports_function_calling() else DEFAULT_MODEL_NAME
    assert agent.func_calling_llm.api_key is not None
    assert agent.func_calling_llm.max_tokens == 4000
    assert agent.func_calling_llm.logprobs == False
    assert agent.func_calling_llm.callbacks == [dummy_func]
    assert isinstance(agent.func_calling_llm, LLM)


def test_agent_with_random_dict_tools():
    def empty_func():
        return "empty function"

    agent = Agent(role="demo", goal="test a tool", tools=[dict(name="tool 1", func=empty_func), ])

    assert [tool._run() == "empty function" for tool in agent.tools]
    assert agent.tools[0].name == "tool 1"


def test_agent_with_custom_tools():
    def send_message(message: str) -> str:
        return message + "_demo"

    class CustomTool(Tool):
        name: str = "custom tool"
        func: Callable[..., Any]

    tool = CustomTool(func=send_message)
    agent = Agent(role="demo", goal="test a tool", tools=[tool])

    assert agent.tools[0] is tool
    assert agent.tools[0]._run(message="hi") == "hi_demo"
    assert agent.tools[0].name == "custom tool"


# @pytest.mark.vcr(filter_headers=["authorization"])
def test_agent_custom_max_iterations():
    from versionhq.task.model import Task

    @tool
    def get_final_answer() -> int:
        """Get the final answer but don't give it yet, just re-use this tool non-stop."""
        return 42

    agent = Agent(role="demo", goal="test goal", maxit=1, allow_delegation=False, tools=[get_final_answer])

    with patch.object(LLM, "call", wraps=LLM(model=DEFAULT_MODEL_NAME).call) as private_mock:
        task = Task(
            description="The final answer is 42. But don't give it yet, instead keep using the `get_final_answer` tool.",
            can_use_agent_tools=True
        )
        agent.execute_task(task=task)
        assert private_mock.call_count == 1


def test_agent_with_knowledge_sources():
    from versionhq.knowledge.source import StringKnowledgeSource
    from versionhq.knowledge.source_docling import DoclingSource
    from versionhq.task.model import Task

    content = "Kuriko's favorite color is gold, and she enjoy Japanese food."
    string_source = StringKnowledgeSource(content=content)
    html = "https://github.blog/security/vulnerability-research/cybersecurity-researchers-digital-detectives-in-a-connected-world/"
    knowledge_sources = [content, string_source, html,]

    agent = Agent(role="Information Agent", goal="Provide information based on knowledge sources", knowledge_sources=knowledge_sources)

    assert  f"{agent.role.replace(' ', '_')}" in agent._knowledge.collection_name
    assert [isinstance(item, StringKnowledgeSource | DoclingSource) for item in agent.knowledge_sources]
    assert agent._knowledge.embedder_config == agent.embedder_config
    assert agent._knowledge.storage and agent._knowledge.storage.embedding_function
    assert agent._knowledge.storage.app is not None
    assert agent._knowledge.storage.collection_name is not None

    task = Task(description="Answer the following question: What is Kuriko's favorite color?")

    with patch("versionhq.knowledge.storage.KnowledgeStorage") as MockKnowledge:
        mock_knowledge_instance = MockKnowledge.return_value
        isinstance(mock_knowledge_instance.sources[0], StringKnowledgeSource)
        mock_knowledge_instance.query.return_value = [{ "content": content }]

        res = task.execute(agent=agent)
        assert "gold" in res.raw.lower()


def test_using_contextual_memory():
    from versionhq.task.model import Task
    from versionhq.task.evaluate import Evaluation
    from versionhq.storage.rag_storage import RAGStorage
    from versionhq.storage.ltm_sqlite_storage import LTMSQLiteStorage

    agent = Agent(role="Researcher", goal="You research about math.", with_memory=True)
    assert agent.short_term_memory.storage and isinstance(agent.short_term_memory.storage, RAGStorage) and agent.short_term_memory.storage.type == "stm"
    assert agent.long_term_memory.storage and isinstance(agent.long_term_memory.storage, LTMSQLiteStorage)

    task = Task(description="Research a topic to teach a kid aged 6 about math.", should_evaluate=True)
    res = task.execute(agent=agent)
    assert isinstance(res.evaluation, Evaluation) and res.evaluation.suggestion_summary and res.evaluation.aggregate_score is not None


def test_disabled_memory():
    from unittest.mock import patch
    from versionhq.task.model import Task
    from versionhq.memory.contextual_memory import ContextualMemory

    agent = Agent(role="Researcher", goal="You research about math.", with_memory=False)
    # assert agent.short_term_memory is None
    # assert agent.long_term_memory is None
    # assert agent.user_memory is None

    task = Task(description="Research a topic to teach a kid aged 6 about math.")

    with patch.object(ContextualMemory, "build_context_for_task") as contextual_mem:
        task.execute(agent=agent)
        contextual_mem.assert_not_called()


def test_agent_with_memory_config():
    from versionhq.storage.ltm_sqlite_storage import LTMSQLiteStorage

    agent_1 = Agent(role="Researcher", goal="You research about math.", with_memory=True, memory_config=dict(provider="mem0"))
    agent_2 = Agent(role="Researcher", goal="You research about math.", with_memory=True, memory_config=dict(provider="mem0", user_id="123"))


    assert agent_1.short_term_memory is not None
    assert agent_1.short_term_memory.memory_provider == "mem0" and agent_1.short_term_memory.storage.memory_type == "stm"
    assert agent_1.long_term_memory and isinstance(agent_1.long_term_memory.storage, LTMSQLiteStorage)
    assert agent_1.user_memory is None

    assert agent_2.short_term_memory is not None
    assert agent_2.short_term_memory.memory_provider == "mem0" and agent_2.short_term_memory.storage.memory_type == "stm"
    assert agent_2.long_term_memory and isinstance(agent_2.long_term_memory.storage, LTMSQLiteStorage)
    assert agent_2.user_memory and agent_2.user_memory.storage and agent_2.user_memory.storage.memory_type == "user"


def test_updating_llm():
    import versionhq as vhq

    agent = vhq.Agent(role="Researcher", goal="You research about math.")
    agent._update_llm(llm="gemini-2.0")
    assert isinstance(agent.llm, vhq.LLM) and "gemini-2.0" in agent.llm.model

    agent._update_llm(llm_config=dict(max_tokens=10000))
    assert agent.llm.max_tokens == 10000

    agent._update_llm(llm="deepseek", llm_config=dict(max_tokens=500))
    assert isinstance(agent.llm, vhq.LLM) and "deepseek-r1" in agent.llm.model and agent.llm.max_tokens == 500


def test_start_with_tools():
    import versionhq as vhq

    def demo_func() -> str:
        return "demo"

    my_tool = vhq.Tool(func=demo_func)

    agent = vhq.Agent(
        role="Tool Handler",
        goal="efficiently use the given tools",
        tools=[my_tool, ]
    )

    res = agent.start(tool_res_as_final=True)
    assert res.tool_output == "demo"



from unittest.mock import patch
import pytest

from versionhq.agent.model import Agent
from versionhq.memory.model import ShortTermMemory, LongTermMemory, MemoryItem, MemoryMetadata, MemoryData



@pytest.fixture
def short_term_memory():
    """Fixture to create a ShortTermMemory instance"""
    agent = Agent(role="Pytest Demo", goal="Search relevant data and provide results")
    return ShortTermMemory(agent=agent)


def test_save_and_search_stm(short_term_memory):
    memory = MemoryItem(
        data=MemoryData(
            agent="test_agent",
            config=dict(val="""test value test value test value test value test value test value test value testvalue test value test value test value test value test value test value cat test value test value test value test value""")
        ),
        metadata=MemoryMetadata(config={ "task": "test_task" })
    )

    with patch.object(ShortTermMemory, "save") as mock_save:
        short_term_memory.save(value=memory.data, metadata=memory.metadata)
        mock_save.assert_called_once_with(value=memory.data, metadata=memory.metadata)

    expected_result = [
        {
            "context": str(memory.data),
            "metadata": { "task": "test_task" },
            "score": 0.95,
        }
    ]
    with patch.object(ShortTermMemory, "search", return_value=expected_result):
        find = short_term_memory.search("test value", score_threshold=0.01)[0]
        assert find["context"] == str(memory.data), "Data value mismatch."
        assert find["metadata"]["task"] == "test_task", "Metadata value mismatch."


@pytest.fixture
def long_term_memory():
    """Fixture to create a LongTermMemory instance"""
    return LongTermMemory()


def test_save_and_search_ltm(long_term_memory):
    memory_data = MemoryData(agent="test_agent", task_description="test_task", task_output="test_output")
    memory_metadata = MemoryMetadata(eval_criteria="test", score=0.5, config={"task": "test_task", "quality": 0.5})

    long_term_memory.save(data=memory_data, metadata=memory_metadata)
    find = long_term_memory.search(query="test_task", latest_n=3)[0]

    assert long_term_memory.storage is not None
    assert find["data"]["agent"] == "test_agent"
    assert find["data"]["task_description"] == "test_task"
    assert find["data"]["task_output"] == "test_output"
    assert find["metadata"]["eval_criteria"] == "test"
    assert find["metadata"]["score"] == 0.5
    assert find["metadata"]["task"] == "test_task"
    assert find["metadata"]["quality"] == "0.5"



import sys
from typing import Type, Any
from pydantic import BaseModel
from pydantic._internal._model_construction import ModelMetaclass
from textwrap import dedent
if 'pydantic.main' not in sys.modules:
    import pydantic.main

sys.modules['pydantic.main'].ModelMetaclass = ModelMetaclass

from versionhq.agent.model import Agent
from versionhq.task.model import ResponseField
from versionhq.task_graph.model import TaskGraph, Task, DependencyType, Node
from versionhq._utils.logger import Logger


def workflow(final_output: Type[BaseModel], context: Any = None, human: bool = False, with_memory: bool = False) -> TaskGraph | None:
    """
    Generate a TaskGraph object to generate the givne final_output most resource-efficiently.
    """

    if not final_output or not isinstance(final_output, ModelMetaclass):
        Logger().log(level="error", message="Missing an expected output in Pydantic model.", color="red")
        return None

    final_output_prompt = ", ".join([k for k in final_output.model_fields.keys()])

    if not final_output_prompt:
        Logger().log(level="error", message="Expected output is in invalid format.", color="red")
        return None

    context_prompt = f'We are designing a resource-efficient workflow using graph algorithm concepts to achieve the following goal: {final_output_prompt}.'

    dep_type_prompt = ", ".join([k for k in DependencyType._member_map_.keys()])

    graph_expert = Agent(
        role="vhq-G Expert",
        goal="design the most resource-efficient workflow graph to achieve the given goal",
        knowledge_sources=[
            "https://en.wikipedia.org/wiki/Graph_theory",
            # "https://www.geeksforgeeks.org/graph-data-structure-and-algorithms/?ref=lbp",
            "https://www.geeksforgeeks.org/graph-and-its-representations/",
            ", ".join([k for k in DependencyType._member_map_.keys()]),
        ],
        llm="gemini-2.0",
        with_memory=with_memory,
        maxit=1,
        max_retry_limit=1,
    )

    task = Task(
        description=dedent(f"Design a resource-efficient workflow to achieve the following goal: {final_output_prompt}. The workflow should consist of a list of detailed tasks that represent decision making points, each with the following information:\nname: A concise name of the task\ndescription: A concise description of the task.\nconnections: A list of target tasks that this task connects to.\ndependency_types: The type of dependency between this task and each of its connected task. \noutput: key output from the task in a word.\n\nUse the following dependency types: {dep_type_prompt}.\n\nPrioritize minimizing resource consumption (computation, memory, and data transfer) when defining tasks, connections, and dependencies.  Consider how data is passed between tasks and aim to reduce unnecessary data duplication or transfer. Explain any design choices made to optimize resource usage."),
        response_fields=[
            ResponseField(title="tasks", data_type=list, items=dict, properties=[
                ResponseField(title="name", data_type=str),
                ResponseField(title="description", data_type=str),
                ResponseField(title="output", data_type=str),
                ResponseField(title="connections", data_type=list, items=str),
                ResponseField(title="dependency_types", data_type=list, items=str),
            ])
        ]
    )
    res = task.execute(agent=graph_expert, context=[context_prompt, context])

    if not res:
        return None

    task_items = res.json_dict["tasks"]
    tasks, nodes = [], []

    for item in task_items:
        key = item["output"].lower().replace(" ", "_") if item["output"] else "output"
        task = Task(name=item["name"], description=item["description"], response_fields=[ResponseField(title=key, data_type=str)])
        tasks.append(task)
        nodes.append(Node(task=task))

    task_graph = TaskGraph(
        nodes={node.identifier: node for node in nodes},
        concl_format=final_output,
        concl=None,
        should_reform=True,
    )

    for res in task_items:
        if res["connections"]:
            dependency_types = [DependencyType[dt] if DependencyType[dt] else DependencyType.FINISH_TO_START for dt in res["dependency_types"]]

            for i, target_task_name in enumerate(res["connections"]):
                source = [v for v in task_graph.nodes.values() if v.task.name == res["name"]][0]
                target = [v for v in task_graph.nodes.values() if v.task.name == target_task_name][0]
                dependency_type = dependency_types[i]
                task_graph.add_dependency(
                    source_node_identifier=source.identifier, target_node_identifier=target.identifier, dependency_type=dependency_type)


    task_graph.visualize()

    if human:
        print('Proceed? Y/n:')
        x = input()

        if x.lower() == "y":
            print("ok. generating agent network")

        else:
            request = input("request?")
            print('ok. regenerating the graph based on your input: ', request)

    return task_graph



import json
import datetime
import sqlite3
from pathlib import Path
from typing import Any, Dict, List, Optional

from versionhq._utils.logger import Logger
from versionhq.storage.utils import fetch_db_storage_path


class LTMSQLiteStorage:
    """
    An updated SQLite storage class for LTM data storage.
    """

    def __init__(self, db_path: Optional[str] = None) -> None:
        if db_path is None:
            db_path = str(Path(fetch_db_storage_path()) / "ltm_storage.db")

        self.db_path = db_path
        self._logger: Logger = Logger(verbose=True)

        Path(self.db_path).parent.mkdir(parents=True, exist_ok=True)
        self._initialize_db()


    def _initialize_db(self):
        """
        Initializes the SQLite database and creates LTM table
        """

        try:
            with sqlite3.connect(self.db_path) as conn:
                cursor = conn.cursor()
                cursor.execute(
                """
                    CREATE TABLE IF NOT EXISTS long_term_memories (
                        id INTEGER PRIMARY KEY AUTOINCREMENT,
                        datetime REAL,
                        data TEXT,
                        metadata TEXT
                    )
                """
                )
                conn.commit()

        except sqlite3.Error as e:
            self._logger.log(level="error", message=f"MEMORY ERROR: An error occurred during database initialization: {str(e)}", color="red")


    def save(self, data: Dict[str, Any] | str, metadata: Optional[Dict[str, Any]] = {}) -> None:
        """
        Saves data to the LTM table with error handling.
        """
        data = data if isinstance(data, dict) else dict(data=data)

        try:
            with sqlite3.connect(self.db_path) as conn:
                cursor = conn.cursor()
                cursor.execute(
            """
                INSERT INTO long_term_memories (datetime, data, metadata)
                VALUES (?, ?, ?)
            """,
                    (datetime.datetime.now().timestamp(), json.dumps(data), json.dumps(metadata)),
                )
                conn.commit()
        except sqlite3.Error as e:
            self._logger.log(level="error", message=f"MEMORY ERROR: An error occurred while saving to LTM: {str(e)}", color="red")


    def load(self, query: str, latest_n: int) -> Optional[List[Dict[str, Any]]]:
        """
        Queries the data row in the storage with error handling.
        """
        try:
            with sqlite3.connect(self.db_path) as conn:
                cursor = conn.cursor()
                cursor.execute(
                f"""
                    SELECT datetime, data, metadata
                    FROM long_term_memories
                    WHERE data LIKE '%{query}%'
                    ORDER BY datetime
                    LIMIT {latest_n}
                """
                )
                rows = cursor.fetchall()
                if rows:
                    return [
                        {
                            "datetime": row[0],
                            "data": json.loads(row[1]),
                            "metadata": json.loads(row[2]),
                        }
                        for row in rows
                    ]

        except sqlite3.Error as e:
            self._logger.log(level="error", message=f"MEMORY ERROR: An error occurred while querying LTM: {e}",color="red")
            return None


    def reset(self) -> None:
        """
        Resets the LTM table with error handling.
        """

        try:
            with sqlite3.connect(self.db_path) as conn:
                cursor = conn.cursor()
                cursor.execute("DELETE FROM long_term_memories")
                conn.commit()

        except sqlite3.Error as e:
            self._logger.log(level="error", message=f"MEMORY ERROR: An error occurred while deleting all rows in LTM: {str(e)}", color="red")
        return None



EVALUATE="""Evaluate the provided task output against the given task description, assigning a score between 0 (worst) and 1 (best) based on the specified criteria. Scores should be numerical (integers or decimals). Provide specific suggestions for improvement. Do not assign identical scores to different criteria:
Task output: {task_output}
Task description: {task_description}
Evaluation criteria: {eval_criteria}
"""



from versionhq.agent.model import Agent
from versionhq.llm.model import DEFAULT_MODEL_NAME

"""
In-house agents to be called across the project.
[Rules] In house agents have names and roles that start with `vhq_`. No customization allowed by client.
"""

vhq_client_manager = Agent(
    role="vhq-Client Manager",
    goal="Efficiently communicate with the client on the task progress",
    llm=DEFAULT_MODEL_NAME,
    with_memory=True,
)


vhq_task_evaluator = Agent(
    role="vhq-Task Evaluator",
    goal="score the output according to the given evaluation criteria.",
    llm=DEFAULT_MODEL_NAME,
    llm_config=dict(top_p=0.8, top_k=30, max_tokens=5000, temperature=0.9),
    maxit=1,
    max_retry_limit=1,
    with_memory=True # refer past eval records of similar tasks
)


vhq_formation_planner = Agent(
    role="vhq-Formation Planner",
    goal="Plan a formation of agents based on the given task descirption.",
    llm="gemini/gemini-2.0-flash-exp",
    llm_config=dict(top_p=0.8, topK=40, temperature=0.9),
    maxit=1,
    max_retry_limit=1,
    knowledge_sources=[
        "Solo is a formation where a single agent with tools, knowledge, and memory handles tasks indivudually. When self-learning mode is on - it will turn into Random formation. Typical usecase is an email agent drafts promo message for the given audience using their own knowledge.",
        "Supervising is a formation where the leader agent gives directions, while sharing its knowledge and memory with subbordinates.Subordinates can be solo agents or networks. Typical usecase is that the leader agent strategizes an outbound campaign plan and assigns components such as media mix or message creation to subordinate agents.",
        "Network is a formation where multple agents can share tasks, knowledge, and memory among network members without hierarchy. Typical usecase is that an email agent and social media agent share the product knowledge and deploy multi-channel outbound campaign. ",
        "Random is a formation where a single agent handles tasks, asking help from other agents without sharing its memory or knowledge. Typical usecase is that an email agent drafts promo message for the given audience, asking insights on tones from other email agents which oversee other customer clusters, or an agent calls the external, third party agent to deploy the campaign. ",
    ]
)


vhq_agent_creator = Agent(
    role="vhq-Agent Creator",
    goal="build an agent that can handle the given task",
    llm="gemini/gemini-2.0-flash-exp",
)



import appdirs
import os
from dotenv import load_dotenv
from pathlib import Path

load_dotenv(override=True)

def fetch_db_storage_path() -> str:
    directory_name = get_project_directory_name()
    data_dir = Path(appdirs.user_data_dir(appname=directory_name, appauthor="Version IO Sdn Bhd", version=None, roaming=False))
    data_dir.mkdir(parents=True, exist_ok=True)
    return str(data_dir)


def get_project_directory_name() -> str:
    """
    Returns the current project directory name
    """
    project_directory_name = os.environ.get("STORAGE_DIR")

    if project_directory_name:
        return project_directory_name
    else:
        cwd = Path.cwd()
        project_directory_name = cwd.name
        return project_directory_name



import datetime

from versionhq.clients.workflow.model import MessagingWorkflow, MessagingComponent
from versionhq.clients.product.model import Product, ProductProvider
from versionhq.tool.composio_tool_vars import ComposioAppName


def test_store_scores():
    """
    Test if the final result will be calcurated using a random subject
    """

    messaging_component = MessagingComponent(message="demo")
    score_raw = 15
    messaging_component.store_scoring_result("demo", score_raw=score_raw)

    assert messaging_component.score is not None
    assert messaging_component.score.result() is not None


def test_score_result():
    messaging_component = MessagingComponent(message="demo")
    score_raw = 15
    messaging_component.store_scoring_result("demo", score_raw=score_raw)

    result = messaging_component.score.result()

    assert result is not None
    assert result != 0


def test_setup_messaging_workflow_with_anonymous_provider():
    product = Product(description="demo p", audience="demo audience", usp="demo usp", landing_url="www.com")
    comp = MessagingComponent(message="demo")
    messaging_workflow = MessagingWorkflow(_created_at=datetime.datetime.now(), messaging_components=[comp, ], product=product)

    assert messaging_workflow.id is not None
    assert messaging_workflow.destination is None


def test_setup_messaging_workflow_with_provider():
    provider = ProductProvider(
        name="demo provider",
        region="US",
        data_pipelines=["data"],
        destination_services=["linkedin", "email",]
    )
    product = Product(
        description="demo p",
        audience="demo audience",
        usp="demo usp",
        landing_url="www.com",
        cohort_timeframe=30,
        provider=provider,
    )
    comp = MessagingComponent(message="demo")
    messaging_workflow = MessagingWorkflow(_created_at=datetime.datetime.now(), messaging_components=[comp, ], product=product)

    assert messaging_workflow.id is not None
    assert messaging_workflow.destination is not None
    assert messaging_workflow.destination in provider.destination_services
    assert isinstance(messaging_workflow.destination, ComposioAppName)



from typing import Any, Callable, List, Optional

from versionhq.tool.model import Tool, ToolSet
from versionhq.tool.tool_handler import ToolHandler
from versionhq.tool.cache_handler import CacheHandler
from versionhq.tool.decorator import tool


def test_create_custom_tool():
    class CustomTool(Tool):
        name: str = "custom tool"
        goal: str = "test a custom tool class"
        func: Callable[..., Any] = lambda x: f"{x}-demo"

    tool = CustomTool()

    assert tool.name == "custom tool"
    assert tool.description == "Tool: custom tool\nArgs: {'x': {'description': '', 'type': 'Any'}}"
    assert tool._run(x="I am a sentence") == "I am a sentence-demo"


def test_run_tool_with_abstract_function():
    def function_to_test(test_list: List) -> int:
        return len(test_list)

    class CustomTool(Tool):
        name: str = "custom tool"
        func: Callable[..., Any] = function_to_test

    tool = CustomTool()
    result = tool._run(["demo 0", "demo 1", "demo 2"])

    assert tool.name == "custom tool"
    assert result == 3


def test_run_tool_with_function():
    def function_to_test(test_list: List) -> int:
        return len(test_list)

    tool = Tool(name="tool", func=function_to_test)
    result = tool._run(test_list=["demo 0", "demo 1", "demo 2"])

    assert tool.name == "tool"
    assert result == 3


def test_setup_cache_function():
    class CustomTool(Tool):
        name: str = "custom tool"
        goal: str = "test a custom tool"
        cache_function: Callable = lambda: False
        func: Callable = lambda sentence: sentence

    tool = CustomTool()
    assert not tool.cache_function()


def test_default_cache_function():
    class CustomTool(Tool):
        name: str = "custom tool"
        func: Callable[..., Any] = lambda x: x

    tool = CustomTool()
    assert tool.cache_function()


def test_tool_annotation():
    @tool("demo")
    def my_tool(test_words: str) -> str:
        """Test a tool decorator."""
        return test_words

    assert my_tool.name == "demo"
    assert my_tool.description == "Tool: demo\nArgs: {'test_words': {'description': '', 'type': 'str'}}"
    assert my_tool.args_schema.model_json_schema() == {'properties': {'test_words': {'title': 'Test Words', 'type': 'string'}}, 'required': ['test_words'], 'title': 'My_ToolSchema', 'type': 'object'}
    assert my_tool.func("testing") == "testing"


def test_tool_handler_turning_off_cache():
    def empty_func():
        return "empty function"

    tool = Tool(name="demo", func=empty_func, should_cache=False)
    tool.run()

    assert isinstance(tool.tool_handler, ToolHandler)
    assert tool.tool_handler.should_cache == False
    assert tool.tool_handler.last_used_tool is None


def test_tool_handler_with_cache():
    def empty_func():
        return "empty function"

    tool = Tool(name="demo", func=empty_func, should_cache=True)
    res = tool.run()

    assert res == "empty function"
    assert isinstance(tool.tool_handler, ToolHandler)
    assert tool.tool_handler.should_cache == True
    assert tool.tool_handler.last_used_tool == ToolSet(tool=tool, kwargs={})
    assert tool.tool_handler.cache.read(tool_name=tool.name, input=str({})) == res


def test_custom_tool_with_cache():
    class CustomTool(Tool):
        name: str = "custom tool"
        func: Optional[Callable] = None

    def demo_func(demo_list: List[Any]) -> int:
        return len(demo_list)

    tool = CustomTool(func=demo_func)
    res = tool.run(params=dict(demo_list=["demo1", "demo2"]))

    assert isinstance(tool.tool_handler, ToolHandler)
    assert tool.tool_handler.should_cache == True
    assert tool.tool_handler.last_used_tool == ToolSet(tool=tool, kwargs=dict(demo_list=["demo1", "demo2"]))



def test_cache_tool():
    my_tool = Tool(
        name="my tool",
        func=lambda x: x + 2,
        cache_handler=CacheHandler()
    )
    res = my_tool._run(x=3)
    my_tool.cache_handler.add(tool_name=my_tool.name, input=str({"x": 3}), output=res)

    assert my_tool.cache_handler.read(tool_name=my_tool.name, input=str({"x": 3})) == 5
    assert my_tool.cache_handler._cache[f"{my_tool.name}-{str({"x": 3})}"] == 5



from abc import ABC, abstractmethod
from typing import Any, Dict, Callable, Type, Optional, get_args, get_origin, get_type_hints
from typing_extensions import Self
from pydantic import InstanceOf, BaseModel, Field, field_validator, model_validator, PrivateAttr, create_model
from pydantic_core import PydanticCustomError

from versionhq.llm.llm_vars import SchemaType
from versionhq.tool.cache_handler import CacheHandler
from versionhq._utils.logger import Logger


class BaseTool(ABC, BaseModel):
    """
    Abstract class for Tool class.
    """

    class ArgsSchemaPlaceholder(BaseModel):
        pass

    _logger: Logger = PrivateAttr(default_factory=lambda: Logger(verbose=True))

    object_type: str = Field(default="function")
    name: str = Field(default=None)
    description: str = Field(default=None)
    properties: Dict[str, Any] = Field(default_factory=dict, description="for llm func calling")
    args_schema: Type[BaseModel] = Field(default_factory=ArgsSchemaPlaceholder)

    tool_handler: Optional[Dict[str, Any] | Any] = Field(default=None, description="store tool_handler to record the tool usage")
    should_cache: bool = Field(default=True, description="whether the tool usage should be cached")
    cache_function: Callable = lambda _args=None, _result=None: True
    cache_handler: Optional[InstanceOf[CacheHandler]] = Field(default=None)


    @field_validator("args_schema", mode="before")
    @classmethod
    def _default_args_schema(cls, v: Type[BaseModel]) -> Type[BaseModel]:
        if not isinstance(v, cls.ArgsSchemaPlaceholder):
            return v

        return type(
            f"{cls.__name__}Schema",
            (BaseModel,),
            { "__annotations__": { k: v for k, v in cls._run.__annotations__.items() if k != "return" }},
        )


    @field_validator("properties", mode="before")
    @classmethod
    def _default_properties(cls, v: Dict[str, Any]) -> Dict[str, Any]:
        p, r = dict(), list()
        for k, v in cls._run.__annotations__.items():
            if k != "return":
                p.update({ k: { "type": SchemaType(type(v)).convert(), "name": k, }} )
                r.append(k)

        return {
            "type": cls.object_type,
            "function": {
                "name": cls.name.replace(" ", "_"),
                "description": cls.description if cls.description else "",
                "parameters": {
                    "type": "object",
                    "properties": p,
                    "required": r,
                    "additionalProperties": False
                },
                "strict": True
            }
        }

    @model_validator(mode="after")
    def set_up_tool_handler(self) -> Self:
        from versionhq.tool.tool_handler import ToolHandler

        if self.tool_handler and not isinstance(self.tool_handler, ToolHandler):
            ToolHandler(**self.tool_handler)

        else:
            self.tool_handler = ToolHandler(cache_handler=self.cache_handler, should_cache=self.should_cache)

        return self

    @abstractmethod
    def _run(self, *args: Any, **kwargs: Any,) -> Any:
        """any handling"""


    @staticmethod
    def _get_arg_annotations(annotation: type[Any] | None) -> str:
        if annotation is None:
            return "None"

        origin = get_origin(annotation)
        args = get_args(annotation)

        if origin is None:
            return annotation.__name__ if hasattr(annotation, "__name__") else str(annotation)

        if args:
            args_str = ", ".join(Tool._get_arg_annotations(arg) for arg in args)
            return f"{origin.__name__}[{args_str}]"

        return origin.__name__


    def _parse_args(self, raw_args: str | dict) -> dict:
        """
        Parse and validate the input arguments against the schema
        """
        if isinstance(raw_args, str):
            try:
                import json
                raw_args = json.loads(raw_args)
            except json.JSONDecodeError as e:
                raise ValueError(f"Failed to parse arguments as JSON: {str(e)}")

        try:
            validated_args = self.args_schema.model_validate(raw_args)
            return validated_args.model_dump()

        except Exception as e:
            raise ValueError(f"Arguments validation failed: {str(e)}")


    def _create_schema(self) -> type[BaseModel]:
        """
        Create a Pydantic schema from a function's signature
        """
        import inspect

        sig = inspect.signature(self.func)
        type_hints = get_type_hints(self.func)
        fields = {}
        for param_name, param in sig.parameters.items():
            if param_name in ("self", "cls"):
                continue

            annotation = type_hints.get(param_name, Any)
            default = ... if param.default == param.empty else param.default
            fields[param_name] = (annotation, Field(default=default))

        schema_name = f"{self.func.__name__.title()}Schema"
        return create_model(schema_name, **fields)



class Tool(BaseTool):
    func: Callable = Field(default=None)


    @model_validator(mode="after")
    def validate_func(self) -> Self:
        if not self.func and not self._run:
            self._logger.log(level="error", message=f"Tool must have a function", color="red")
            raise PydanticCustomError("function_missing", f"Function is missing in the tool.", {})

        elif self.func and not isinstance(self.func, Callable):
            self._logger.log(level="error", message=f"The tool is missing a valid function", color="red")
            raise PydanticCustomError("invalid_function", f"The value in the function field must be callable.", {})

        else:
            try:
                self.args_schema = self._create_schema_from_function()
                self._validate_function_signature()

            except Exception as e:
                self._logger.log(level="error", message=f"The tool is missing a valid function: {str(e)}", color="red")
                raise PydanticCustomError("invalid_function", f"Invalid function: {str(e)}", {})

        return self


    @model_validator(mode="after")
    def set_up_name(self) -> Self:
        if not self.name:
            self.name = self.func.__name__ if self.func.__name__ != "<lambda>" else "random_func"

        return self


    @model_validator(mode="after")
    def set_up_description(self) -> Self:
        if not self.description:
            if not self.args_schema:
                self.args_schema = self._default_args_schema(self)

            args_schema = {
                name: {
                    "description": field.description if field.description else "",
                    "type": self._get_arg_annotations(field.annotation),
                }
                for name, field in self.args_schema.model_fields.items()
            }
            self.description = f"Tool: {self.name}\nArgs: {args_schema}"

        return self


    @model_validator(mode="after")
    def set_up_args_schema(self) -> Self:
        """
        Set up args schema based on the given function.
        """
        if self.func:
            self.args_schema = self._create_schema_from_function()
        return self


    @model_validator(mode="after")
    def set_up_func_calling_properties(self) -> Self:
        """
        Format function_calling params from args_schema.
        """

        p, r = dict(), list()
        if not self.args_schema:
            self.args_schema = self.set_up_args_schema()

        for name, field in self.args_schema.model_fields.items():
            if name != "kwargs" and name != "args":
                p.update(
                    {
                        name: {
                            "description": field.description if field.description else "",
                            "type": SchemaType(self._get_arg_annotations(field.annotation)).convert()
                        }
                    }
                )
                r.append(name)

        properties = {
            "type": self.object_type,
            "function": {
                "name": self.name.replace(" ", "_"),
                "description": self.description if self.description else "a tool function to execute",
                "parameters": {
                    "type": "object",
                    "properties": p,
                    "required": r,
                    "additionalProperties": False
                },
                "strict": True,
            },
        }
        self.properties = properties
        return self


    def _create_schema_from_function(self) -> type[BaseModel]:
        """
        Create a Pydantic schema from a function's signature
        """
        import inspect
        sig = inspect.signature(self.func)
        type_hints = get_type_hints(self.func)
        fields = {}
        for param_name, param in sig.parameters.items():
            if param_name in ("self", "cls"):
                continue

            annotation = type_hints.get(param_name, Any)
            default = ... if param.default == param.empty else param.default
            fields[param_name] = (annotation, Field(default=default))

        schema_name = f"{self.func.__name__.title()}Schema"
        return create_model(schema_name, **fields)


    def _validate_function_signature(self) -> None:
        """
        Validate that the function signature matches the args schema.
        """

        import inspect
        sig = inspect.signature(self.func)
        schema_fields = self.args_schema.model_fields

        for param_name, param in sig.parameters.items():
            if param_name in ("self", "cls"):
                continue

            if param.kind in (inspect.Parameter.VAR_KEYWORD, inspect.Parameter.VAR_POSITIONAL):
                continue

            if param.default == inspect.Parameter.empty:
                if param_name not in schema_fields:
                    raise ValueError(f"Required function parameter '{param_name}' not found in args_schema")


    def _run(self, *args: Any, **kwargs: Any) -> Any:
        return self.func(*args, **kwargs)


    def _handle_toolset(self, params: Dict[str, Any] = None) -> Any:
        """
        Return cached results or run the function and record the results.
        """

        from versionhq.tool.tool_handler import ToolHandler

        if not self.args_schema:
            self.args_schema = self._create_schema_from_function()

        result = None
        acceptable_args = self.args_schema.model_json_schema()["properties"].keys()
        acceptable_kwargs = { k: v for k, v in params.items() if k in acceptable_args } if params else dict()
        parsed_kwargs = self._parse_args(raw_args=acceptable_kwargs)
        tool_set = ToolSet(tool=self, kwargs=acceptable_kwargs)

        if not self.tool_handler or not isinstance(self.tool_handler, ToolHandler):
            self.tool_handler = ToolHandler(last_used_tool=tool_set, cache_handler=self.cache_handler, should_cache=self.should_cache)

        try:
            if self.tool_handler.has_called_before(tool_set) or self.tool_handler.cache:
                result = self.tool_handler.cache.read(tool_name=tool_set.tool.name, input=str(tool_set.kwargs))

            if not result:
                result = self.func(**parsed_kwargs)

            if self.should_cache is True:
                self.tool_handler.record_last_tool_used(last_used_tool=tool_set, output=result, should_cache=self.should_cache)

            return result

        except:
            self.tool_handler.error = "Agent error"
            return result


    def run(self, params: Dict[str, Any] = None) -> Any:
        """
        Execute a tool using a toolset and cached tools
        """
        result = self._handle_toolset(params)
        return result


class ToolSet(BaseModel):
    """
    Store the tool called and any kwargs used. (The tool name and kwargs will be stored in the cache.)
    """
    tool: InstanceOf[Tool]| Type[Tool] = Field(..., description="store the tool instance to be called.")
    kwargs: Optional[Dict[str, Any]] = Field(..., description="kwargs passed to the tool")



"""Test code snippets on docs/core/task.md"""


def test_docs_core_task_a():
    import versionhq as vhq

    task = vhq.Task(description="MY AMAZING TASK")

    import uuid
    assert uuid.UUID(str(task.id), version=4)


def test_docs_core_task_b():
    import versionhq as vhq

    task = vhq.Task(description="MY AMAZING TASK")
    res = task.execute()

    assert isinstance(res, vhq.TaskOutput)
    assert res.raw and res.json
    assert task.processed_agents is not None


def test_docs_core_task_c():
    from pydantic import BaseModel
    from typing import Any

    class Demo(BaseModel):
        """
        A demo pydantic class to validate the outcome with various nested data types.
        """
        demo_1: int
        demo_2: float
        demo_3: str
        demo_4: bool
        demo_5: list[str]
        demo_6: dict[str, Any]
        demo_nest_1: list[dict[str, Any]] # 1 layer of nested child is ok.
        demo_nest_2: list[list[str]]
        demo_nest_3: dict[str, list[str]]
        demo_nest_4: dict[str, dict[str, Any]]
        # error_1: list[list[dict[str, list[str]]]] # <- Error due to 2+ layers of nested children.
        # error_2: InstanceOf[AnotherPydanticClass] # <- Error due to non-typing annotation.

    # 2. Define a task
    import versionhq as vhq
    task = vhq.Task(
        description="generate random output that strictly follows the given format",
        pydantic_output=Demo,
    )

    # 3. Execute
    res = task.execute()

    assert isinstance(res, vhq.TaskOutput) and res.task_id is task.id
    assert isinstance(res.raw, str) and isinstance(res.json_dict, dict)
    assert [
        getattr(res.pydantic, k) and v.annotation == Demo.model_fields[k].annotation
        for k, v in res.pydantic.model_fields.items()
    ]


def test_docs_core_task_d():
    import versionhq as vhq

    # 1. Define a list of ResponseField objects
    demo_response_fields = [
        vhq.ResponseField(title="demo_1", data_type=int),
        vhq.ResponseField(title="demo_2", data_type=float),
        vhq.ResponseField(title="demo_3", data_type=str),
        vhq.ResponseField(title="demo_4", data_type=bool),
        vhq.ResponseField(title="demo_5", data_type=list, items=str),
        vhq.ResponseField(title="demo_6", data_type=dict, properties=[vhq.ResponseField(title="demo-item", data_type=str)]),
        vhq.ResponseField(title="demo_nest_1", data_type=list, items=dict, properties=([
            vhq.ResponseField(title="nest1", data_type=dict, properties=[vhq.ResponseField(title="nest11", data_type=str)])
        ])),
        vhq.ResponseField(title="demo_nest_2", data_type=list, items=list),
        vhq.ResponseField(title="demo_nest_3", data_type=dict, properties=[
            vhq.ResponseField(title="nest1", data_type=list, items=str)
        ]),
        vhq.ResponseField(title="demo_nest_4", data_type=dict, properties=[
            vhq.ResponseField(title="nest1", data_type=dict, properties=[vhq.ResponseField(title="nest12", data_type=str)])
        ])
    ]

    task = vhq.Task(
        description="Output random values strictly following the data type defined in the given response format.",
        response_fields=demo_response_fields
    )

    res = task.execute()

    assert isinstance(res, vhq.TaskOutput) and res.task_id is task.id
    assert res.raw and res.json and res.pydantic is None
    assert [v and type(v) == task.response_fields[i].data_type for i, (k, v) in enumerate(res.json_dict.items())]


def test_docs_core_task_e():
    import versionhq as vhq
    from pydantic import BaseModel
    from typing import Any

    # 1. Define and execute a sub task with Pydantic output.
    class Sub(BaseModel):
        sub1: list[dict[str, Any]]
        sub2: dict[str, Any]

    sub_task = vhq.Task(
        description="generate random values that strictly follows the given format.",
        pydantic_output=Sub
    )
    sub_res = sub_task.execute()

    # 2. Define a main task, callback function to format the final response.
    class Main(BaseModel):
        main1: list[Any] # <= assume expecting to store Sub object.
        main2: dict[str, Any]

    def format_response(sub, main1, main2) -> Main:
        main1.append(sub)
        main = Main(main1=main1, main2=main2)
        return main

    # 3. Execute
    main_task = vhq.Task(
        description="generate random values that strictly follows the given format.",
        pydantic_output=Main,
        callback=format_response,
        callback_kwargs=dict(sub=Sub(sub1=sub_res.pydantic.sub1, sub2=sub_res.pydantic.sub2)),
    )
    res = main_task.execute(context=sub_res.raw) # [Optional] Adding sub_task's response as context.

    assert [item for item in res.callback_output.main1 if isinstance(item, Sub)]


def test_docs_core_task_f():
    import versionhq as vhq

    sub_task_1 = vhq.Task(description="Run a sub demo part 1")
    sub_res = sub_task_1.execute()
    sub_task_2 = vhq.Task(description="Run a sub demo part 2")
    task = vhq.Task(description="Run a main demo")
    # res = task.execute(context=[sub_res, sub_task_2, "context to add in string",])
    task_prompt = task._prompt(context=[sub_res, sub_task_2, "context to add in string"])

    assert sub_res.to_context_prompt() in task_prompt
    assert sub_task_2.output and sub_task_2.output.to_context_prompt() in task_prompt
    assert "context to add in string" in task_prompt


def test_docs_core_task_g():
    import versionhq as vhq
    task = vhq.Task(
        description="return the output following the given prompt.",
        allow_delegation=True
    )
    task.execute()

    assert task.output is not None
    assert "vhq-Delegated-Agent" in task.processed_agents
    assert task.delegations ==1


def test_docs_core_task_h():
    import versionhq as vhq

    task = vhq.Task(description="Return a word: 'test'", type=vhq.TaskExecutionType.ASYNC)

    from unittest.mock import patch
    with patch.object(vhq.Agent, "execute_task", return_value="test") as execute:
        res = task.execute()
        assert res.raw == "test"
        execute.assert_called_once_with(task=task, context=None, task_tools=list())


def test_docs_core_task_i():
    import versionhq as vhq

    def random_func(message: str) -> str:
        return message + "_demo"

    tool = vhq.Tool(name="tool", func=random_func)
    tool_set = vhq.ToolSet(tool=tool, kwargs=dict(message="empty func"))
    task = vhq.Task(
        description="execute the given tools",
        tools=[tool_set,],
        tool_res_as_final=True
    )

    res = task.execute()
    assert res.tool_output == "empty func_demo"


def test_docs_core_task_j():
    import versionhq as vhq

    simple_tool = vhq.Tool(name="simple tool", func=lambda x: "simple func")
    agent = vhq.Agent(role="demo", goal="execute tools", tools=[simple_tool,])
    task = vhq.Task(
        description="execute tools",
        can_use_agent_tools=True,
        tool_res_as_final=True
    )
    res = task.execute(agent=agent)
    assert res.tool_output == "simple func"


def test_docs_core_task_k():
    import versionhq as vhq

    def callback_func(condition: str, test1: str):
        return f"Result: {test1}, condition added: {condition}"

    task = vhq.Task(
        description="return the output following the given prompt.",
        callback=callback_func,
        callback_kwargs=dict(condition="demo for pytest")
    )
    res = task.execute()

    assert res and isinstance(res, vhq.TaskOutput)
    assert res.task_id is task.id
    assert "demo for pytest" in res.callback_output



from typing import Any, Dict, Type
from pydantic import BaseModel


def process_config(values_to_update: Dict[str, Any], model_class: Type[BaseModel]) -> Dict[str, Any]:
    """
    Process the config dictionary and update the values accordingly.
    Refer to the Pydantic model class for field validation.
    """

    config = values_to_update.pop("config") if "config" in values_to_update else {}

    if config:
        for k, v in config.items():
            if k not in model_class.model_fields or values_to_update.get(k) is not None:
                continue

            if isinstance(v, dict) and isinstance(values_to_update.get(k), dict):
                values_to_update[k].update(v)
            else:
                values_to_update[k] = v

    return values_to_update
